"""Logging and log related utilities."""
from __future__ import annotations

from .formatters import console_log_formatter, json_formatter, key_value_formatter
from .job_logging_service import (
    JobLoggingService,
    MissingJobLogException,
    SizeThresholdJobLogException,
)
from .output_logger import OutputLogger
from .utils import DEFAULT_LEVEL, LEVELS, capture_subprocess_output, setup_logging



"""Defines DBT-specific plugins."""
from __future__ import annotations

import logging
from pathlib import Path

from meltano.core.error import PluginInstallError
from meltano.core.plugin import BasePlugin, PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.setting_definition import SettingDefinition, SettingKind
from meltano.core.transform_add_service import TransformAddService

logger = logging.getLogger(__name__)


class DbtInvoker(PluginInvoker):
    def Popen_options(self):  # noqa: N802
        return {**super().Popen_options(), "cwd": self.plugin_config["project_dir"]}


class DbtPlugin(BasePlugin):
    __plugin_type__ = PluginType.TRANSFORMERS

    invoker_class = DbtInvoker


class DbtTransformPluginInstaller:
    def __init__(self, project, plugin: ProjectPlugin):
        self.project = project
        self.plugin = plugin

    async def install(self, reason, clean):
        """Install the transform into the project."""
        if reason not in {PluginInstallReason.ADD, PluginInstallReason.UPGRADE}:
            logger.info(
                f"Run `meltano add transform {self.plugin.name}` to re-add this transform to your dbt project."
            )
            return

        try:
            transform_add_service = TransformAddService(self.project)

            # Add repo to my-test-project/transform/packages.yml
            packages_path = transform_add_service.packages_file.relative_to(
                self.project.root
            )
            logger.info(
                f"Adding dbt package '{self.plugin.pip_url}' to '{packages_path}'..."
            )
            transform_add_service.add_to_packages(self.plugin)

            # Add model and vars to my-test-project/transform/dbt_project.yml
            dbt_project_path = transform_add_service.dbt_project_file.relative_to(
                self.project.root
            )
            logger.info(f"Adding dbt model to '{dbt_project_path}'...")
            transform_add_service.update_dbt_project(self.plugin)
        except PluginNotFoundError:
            raise PluginInstallError(
                "Transformer 'dbt' is not installed. Run `meltano add transformer dbt` to add it to your project."
            )
        except FileNotFoundError as err:
            relative_path = Path(err.filename).relative_to(self.project.root)
            raise PluginInstallError(
                f"File {relative_path}' could not be found. Run `meltano add files files-dbt` to set up a dbt project."
            )


class DbtTransformPlugin(BasePlugin):
    __plugin_type__ = PluginType.TRANSFORMS

    installer_class = DbtTransformPluginInstaller

    EXTRA_SETTINGS = [
        SettingDefinition(name="_package_name", value="$MELTANO_TRANSFORM_NAMESPACE"),
        SettingDefinition(name="_vars", kind=SettingKind.OBJECT, value={}),
    ]

    def is_invokable(self):
        return False



"""Starts WSGI Webserver that will run the API App for a Meltano Project."""
from __future__ import annotations

import logging
import platform
import threading

from meltano.core.meltano_invoker import MeltanoInvoker
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.settings_service import FeatureFlags
from meltano.core.utils.pidfile import PIDFile


class APIWorker(threading.Thread):
    """The Base APIWorker Class."""

    def __init__(self, project: Project, reload=False):
        """Initialize the API Worker class with the project config.

        Args:
            project: Project class.
            reload: Boolean.
        """
        super().__init__()

        self.project = project
        self.reload = reload
        self.pid_file = PIDFile(self.project.run_dir("gunicorn.pid"))
        self.settings_service = ProjectSettingsService(self.project.find())

    def run(self):
        """Run the initalized API Workers with the App Server requested."""
        with self.settings_service.feature_flag(
            FeatureFlags.ENABLE_UVICORN, raise_error=False
        ) as allow:

            enable_uvicorn = allow

            # Use Uvicorn when on Windows
            if platform.system() == "Windows":
                if enable_uvicorn:
                    logging.debug("ff.enable_uvicorn enabled, starting uvicorn.")
                else:
                    logging.warning(
                        "Windows OS detected auto setting ff.enable_uvicorn"
                    )
                    logging.warning(
                        "Add ff.enable_uvicorn: True to your meltano.yml to supress this waring"
                    )
                    enable_uvicorn = True

            # Start uvicorn to serve API and Ui
            if enable_uvicorn:
                settings_for_apiworker = self.settings_service

                arg_bind_host = str(settings_for_apiworker.get("ui.bind_host"))
                arg_bind_port = str(settings_for_apiworker.get("ui.bind_port"))
                arg_loglevel = str(settings_for_apiworker.get("cli.log_level"))
                arg_forwarded_allow_ips = str(
                    settings_for_apiworker.get("ui.forwarded_allow_ips")
                )

                # If windows and 127.0.0.1 only allowed changing bind host to accomidate
                if platform.system() == "Windows":
                    if (
                        arg_forwarded_allow_ips == "127.0.0.1"
                        and arg_bind_host == "0.0.0.0"  # noqa: S104
                    ):
                        # If left at 0.0.0.0 the server will respond to any request receieved on any interface
                        arg_bind_host = "127.0.0.1"

                # Setup args for uvicorn using bind info from the project setings service
                args = [
                    "--host",
                    arg_bind_host,
                    "--port",
                    arg_bind_port,
                    "--loop",
                    "asyncio",
                    "--interface",
                    "wsgi",
                    "--log-level",
                    arg_loglevel,
                    "--forwarded-allow-ips",
                    arg_forwarded_allow_ips,
                    "--timeout-keep-alive",
                    "600",
                ]

                # Add reload argument if reload is true
                if self.reload:

                    args += [
                        "--reload",
                    ]

                # Add the Meltano API app, factory create_app function combo to the args
                args += [
                    "--factory",
                    "meltano.api.app:create_app",
                ]

                # Start uvicorn using the MeltanoInvoker
                MeltanoInvoker(self.project).invoke(args, command="uvicorn")

            else:
                # Use Gunicorn when feature flag start_uvicorn is not set

                args = [
                    "--config",
                    "python:meltano.api.wsgi",
                    "--pid",
                    str(self.pid_file),
                ]

                if self.reload:
                    args += ["--reload"]

                args += ["meltano.api.app:create_app()"]

                MeltanoInvoker(self.project).invoke(args, command="gunicorn")

    def pid_path(self):
        """Give the path name of the projects gunicorn.pid file location.

        Returns:
            Path object that gives the direct locationo of the gunicorn.pid file.
        """
        return self.project.run_dir("gunicorn.pid")

    def stop(self):
        """Terminnate active gunicorn workers that have placed a PID in the project's gunicorn.pid file."""
        self.pid_file.process.terminate()



from __future__ import annotations

from typing import Callable


class visit_with:  # noqa: N801
    def __init__(self, visit: Callable):
        self.visit = visit

    def __call__(self, base_cls):
        class Visitor(base_cls):
            def visit(inner_self, node, *args, **kwargs):  # noqa: N805
                return self.visit(node, inner_self, *args, **kwargs)

        return Visitor



"""Project context for the Snowplow tracker."""

from __future__ import annotations

import uuid
from enum import Enum, auto

from cached_property import cached_property
from snowplow_tracker import SelfDescribingJson
from structlog.stdlib import get_logger

from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.tracking.schemas import ProjectContextSchema
from meltano.core.utils import hash_sha256

logger = get_logger(__name__)


class ProjectUUIDSource(Enum):
    """The source of the `project_uuid` used for telemetry."""

    # The UUID was explicitly provided in the config as the `project_id`.
    explicit = auto()

    # The UUID was derived by hashing the `project_id` in the config.
    derived = auto()

    # The UUID was randomly generated (UUID v4) since no `project_id` was configured.
    random = auto()


class ProjectContext(SelfDescribingJson):
    """Tracking context for the Meltano project."""

    def __init__(self, project: Project, client_id: uuid.UUID):
        """Initialize a meltano tracking "project" context.

        Args:
            project: The Meltano project.
            client_id: The client ID from `analytics.json`.
        """
        self.project = project
        self.settings_service = ProjectSettingsService(project)
        (
            send_anonymous_usage_stats,
            send_anonymous_usage_stats_metadata,
        ) = self.settings_service.get_with_metadata("send_anonymous_usage_stats")

        super().__init__(
            ProjectContextSchema.url,
            {
                "context_uuid": str(uuid.uuid4()),
                "project_uuid": str(self.project_uuid),
                "project_uuid_source": self.project_uuid_source.name,
                "client_uuid": str(client_id),
                "environment_name_hash": (
                    hash_sha256(self.project.active_environment.name)
                    if self.project.active_environment
                    else None
                ),
                "send_anonymous_usage_stats": send_anonymous_usage_stats,
                "send_anonymous_usage_stats_source": (
                    send_anonymous_usage_stats_metadata["source"].value
                ),
            },
        )

    @property
    def project_uuid_source(self) -> ProjectUUIDSource:
        """Obtain the source of the `project_uuid` used for telemetry.

        Returns:
            ProjectUUIDSource: The source of the `project_uuid` used for telemetry.
        """
        # Ensure the `project_uuid` has been generated
        self.project_uuid  # noqa: WPS428
        return self._project_uuid_source

    @cached_property
    def project_uuid(self) -> uuid.UUID:
        """Obtain the `project_id` from the project config file.

        If it is not found (e.g. first time run), generate a valid v4 UUID, and and store it in the
        project config file.

        Returns:
            The project UUID.
        """
        project_id_str = self.settings_service.get("project_id")

        if project_id_str:
            try:
                # Project ID might already be a UUID
                project_id = uuid.UUID(project_id_str)
            except ValueError:
                # If the project ID is not a UUID, then we hash it, and use the hash to make a UUID
                project_id = uuid.UUID(hash_sha256(project_id_str)[::2])
                self._project_uuid_source = ProjectUUIDSource.derived
            else:
                self._project_uuid_source = ProjectUUIDSource.explicit
        else:
            project_id = uuid.uuid4()
            self._project_uuid_source = ProjectUUIDSource.random

        return project_id



"""Base Error classes."""

from __future__ import annotations

from asyncio.streams import StreamReader
from asyncio.subprocess import Process
from enum import Enum


class ExitCode(int, Enum):  # noqa: D101
    OK = 0
    FAIL = 1
    NO_RETRY = 2


class Error(Exception):
    """Base exception for ELT errors."""

    def exit_code(self):  # noqa: D102
        return ExitCode.FAIL


class ExtractError(Error):
    """Error in the extraction process, like API errors."""

    def exit_code(self):  # noqa: D102
        return ExitCode.NO_RETRY


class AsyncSubprocessError(Exception):
    """Happens when an async subprocess exits with a resultcode != 0."""

    def __init__(
        self,
        message: str,
        process: Process,
        stderr: str | None = None,
    ):  # noqa: DAR101
        """Initialize AsyncSubprocessError."""
        self.process = process
        self._stderr: str | StreamReader | None = stderr or process.stderr
        super().__init__(message)

    @property
    async def stderr(self) -> str | None:
        """Return the output of the process to stderr."""
        if not self._stderr:  # noqa: DAR201
            return None
        elif not isinstance(self._stderr, str):
            stream = await self._stderr.read()
            self._stderr = stream.decode("utf-8")

        return self._stderr


class PluginInstallError(Exception):
    """Exception for when a plugin fails to install."""


class PluginInstallWarning(Exception):
    """Exception for when a plugin optional optional step fails to install."""


class EmptyMeltanoFileException(Exception):
    """Exception for empty meltano.yml file."""



"""Extractor selection management CLI."""
from __future__ import annotations

from contextlib import closing

import click

from meltano.core.db import project_engine
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin.singer.catalog import SelectionType, SelectPattern
from meltano.core.project import Project
from meltano.core.select_service import SelectService
from meltano.core.utils import click_run_async

from . import cli
from .params import pass_project
from .utils import CliError, InstrumentedCmd


def selection_color(selection):
    """Return the appropriate colour for given SelectionType."""
    if selection is SelectionType.SELECTED:
        return "bright_green"
    elif selection is SelectionType.AUTOMATIC:
        return "bright_white"
    elif selection is SelectionType.EXCLUDED:
        return "red"


def selection_mark(selection):
    """
    Return the mark to indicate the selection type of an attribute.

    Examples:
      [automatic]
      [selected ]
      [excluded ]
    """
    colwidth = max(map(len, SelectionType))  # size of the longest mark
    return f"[{selection:<{colwidth}}]"


@cli.command(cls=InstrumentedCmd, short_help="Manage extractor selection patterns.")
@click.argument("extractor")
@click.argument("entities_filter", default="*")
@click.argument("attributes_filter", default="*")
@click.option("--list", is_flag=True, help="List the current selected tap attributes.")
@click.option(
    "--all",
    is_flag=True,
    help="Show all the tap attributes with their selected status.",
)
@click.option(
    "--rm",
    "--remove",
    "remove",
    is_flag=True,
    help="Remove previously added select patterns.",
)
@click.option(
    "--exclude",
    is_flag=True,
    help="Exclude all attributes that match specified pattern.",
)
@pass_project(migrate=True)
@click.pass_context
@click_run_async
async def select(
    ctx: click.Context,
    project: Project,
    extractor: str,
    entities_filter: str,
    attributes_filter: str,
    **flags: dict[str, bool],
):
    """
    Manage extractor selection patterns.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#select
    """
    try:
        if flags["list"]:
            await show(project, extractor, show_all=flags["all"])
        else:
            update(
                project,
                extractor,
                entities_filter,
                attributes_filter,
                exclude=flags["exclude"],
                remove=flags["remove"],
            )
    except PluginExecutionError as err:
        raise CliError(f"Cannot list the selected attributes: {err}") from err


def update(
    project, extractor, entities_filter, attributes_filter, exclude=False, remove=False
):
    """Update select pattern for a specific extractor."""
    select_service = SelectService(project, extractor)
    select_service.update(entities_filter, attributes_filter, exclude, remove)


async def show(project, extractor, show_all=False):
    """Show selected."""
    _, Session = project_engine(project)  # noqa: N806
    select_service = SelectService(project, extractor)

    with closing(Session()) as session:
        list_all = await select_service.list_all(session)

    # legend
    click.secho("Legend:")
    for sel_type in SelectionType:
        click.secho(f"\t{sel_type}", fg=selection_color(sel_type))

    # report
    click.secho("\nEnabled patterns:")

    select_pattern: SelectPattern
    for select_pattern in map(SelectPattern.parse, select_service.current_select):
        color = "red" if select_pattern.negated else "white"
        click.secho(f"\t{select_pattern.raw}", fg=color)

    click.secho("\nSelected attributes:")
    for stream, prop in tuple(
        (strm, prp)
        for strm in sorted(list_all.streams)
        for prp in sorted(list_all.properties[strm.key])
    ):
        entry_selection = stream.selection + prop.selection
        mark = selection_mark(entry_selection)
        if show_all or entry_selection:
            click.secho(f"\t{mark} ", fg=selection_color(entry_selection), nl=False)
            click.secho(stream.key, fg=selection_color(stream.selection), nl=False)
            click.secho(f".{prop.key}", fg=selection_color(entry_selection))



"""Various utils and formatters for log rendering control."""

from __future__ import annotations

from typing import Sequence

import structlog
from structlog.types import Processor

TIMESTAMPER = structlog.processors.TimeStamper(fmt="iso")

LEVELED_TIMESTAMPED_PRE_CHAIN = frozenset(
    (
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_log_level,
        TIMESTAMPER,
    )
)


def _process_formatter(processor: Processor):
    """Use _process_formatter to configure a structlog.stdlib.ProcessFormatter.

    It will automatically add log level and timestamp fields to any log entries not originating from structlog.

    Args:
        processor: A structlog message processor such as structlog.dev.ConsoleRenderer.
    """
    return structlog.stdlib.ProcessorFormatter(
        processor=processor, foreign_pre_chain=LEVELED_TIMESTAMPED_PRE_CHAIN
    )


def console_log_formatter(colors: bool = False) -> None:
    """Create a logging formatter for console rendering that supports colorization."""
    return _process_formatter(structlog.dev.ConsoleRenderer(colors=colors))


def key_value_formatter(
    sort_keys: bool = False,
    key_order: Sequence[str] | None = None,
    drop_missing: bool = False,
) -> None:
    """Create a logging formatter that renders lines in key=value format."""
    return _process_formatter(
        processor=structlog.processors.KeyValueRenderer(
            sort_keys=sort_keys, key_order=key_order, drop_missing=drop_missing
        )
    )


def json_formatter() -> None:
    """Create a logging formatter that renders lines in JSON format."""
    return _process_formatter(processor=structlog.processors.JSONRenderer())



from __future__ import annotations

import os

from meltano.api.headers import JSON_SCHEME_HEADER, VERSION_HEADER
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import truthy

# Flask
# -----------------
THREADS_PER_PAGE = 1
PROFILE = truthy(os.getenv("FLASK_PROFILE"))


# Meltano
# -----------------
MELTANO_UI_URL = os.getenv("MELTANO_UI_URL", "/")

API_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_FOLDER = os.path.join(API_ROOT_DIR, "static/tmp")
PROJECT_ROOT_DIR = os.path.dirname(API_ROOT_DIR)

# Flask-SQLAlchemy
# -----------------
SQLALCHEMY_ECHO = False
SQLALCHEMY_TRACK_MODIFICATIONS = False

# Flask-Security
# -----------------

SECURITY_PASSWORD_HASH = "bcrypt"  # noqa: S105
SECURITY_REGISTERABLE = False
SECURITY_CHANGEABLE = True
SECURITY_RECOVERABLE = False
SECURITY_CONFIRMABLE = False
SECURITY_URL_PREFIX = "/auth"
SECURITY_USER_IDENTITY_ATTRIBUTES = ("username", "email")
SECURITY_SEND_REGISTER_EMAIL = False
SECURITY_SEND_PASSWORD_CHANGE_EMAIL = False

SECURITY_MSG_USERNAME_NOT_PROVIDED = ("You must provide a username.", "error")
SECURITY_MSG_USERNAME_INVALID = (
    "Username must be at least 6 characters long.",
    "error",
)
SECURITY_MSG_USERNAME_ALREADY_TAKEN = ("This username is already taken.", "error")

# Flask-RESTful
# -----------------

RESTFUL_JSON = {}

# Flask-Executor
# -----------------

EXECUTOR_PROPAGATE_EXCEPTIONS = True


# Flask-CORS
# -----------------

CORS_EXPOSE_HEADERS = [VERSION_HEADER]
CORS_ALLOW_HEADERS = ["CONTENT-TYPE", JSON_SCHEME_HEADER]


class ProjectSettings:
    settings_map = {
        "SERVER_NAME": "ui.server_name",
        "SESSION_COOKIE_DOMAIN": "ui.session_cookie_domain",
        "SESSION_COOKIE_SECURE": "ui.session_cookie_secure",
        "SECRET_KEY": "ui.secret_key",
        # Flask-Security
        "SECURITY_PASSWORD_SALT": "ui.password_salt",
        # Flask-SQLAlchemy
        "SQLALCHEMY_DATABASE_URI": "database_uri",
        # Flask-Authlib
        "GITLAB_CLIENT_ID": "oauth.gitlab.client_id",
        "GITLAB_CLIENT_SECRET": "oauth.gitlab.client_secret",
        # Flask-Mail
        "MAIL_SERVER": "mail.server",
        "MAIL_PORT": "mail.port",
        "MAIL_DEFAULT_SENDER": "mail.default_sender",
        "MAIL_USE_TLS": "mail.use_tls",
        "MAIL_USERNAME": "mail.username",
        "MAIL_PASSWORD": "mail.password",
        "MAIL_DEBUG": "mail.debug",
    }

    def __init__(self, project: Project):
        self.settings_service = ProjectSettingsService(project)

    def as_dict(self):
        return {
            config_key: self.settings_service.get(setting_name)
            for config_key, setting_name in self.settings_map.items()
        }



"""Plugin Add CLI."""

from __future__ import annotations

import click

from meltano.cli import cli
from meltano.cli.params import pass_project
from meltano.cli.utils import (
    CliError,
    PartialInstrumentedCmd,
    add_plugin,
    add_required_plugins,
    check_dependencies_met,
    install_plugins,
)
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.tracking import CliEvent, PluginsTrackingContext, Tracker


@cli.command(  # noqa: WPS238
    cls=PartialInstrumentedCmd,
    short_help="Add a plugin to your project.",
)
@click.argument("plugin_type", type=click.Choice(PluginType.cli_arguments()))
@click.argument("plugin_name", nargs=-1, required=True)
@click.option(
    "--inherit-from",
    help=(
        "Add a plugin inheriting from an existing plugin in the project"
        + " or a discoverable plugin identified, by name."
    ),
)
@click.option(
    "--variant",
    help="Add a specific (non-default) variant of the identified discoverable plugin.",
)
@click.option(
    "--as",
    "as_name",
    help=(
        "Shorthand for '--inherit-from', that can be used to add a discoverable "
        + "plugin to your project with a different name. "
        + "Usage:\b\n\nadd <type> <inherit-from> --as <name>"
    ),
)
@click.option(
    "--custom",
    is_flag=True,
    help="Add a custom plugin. The command will prompt you for the package's base plugin description metadata.",
)
@click.option(
    "--no-install",
    is_flag=True,
    help="Do not install the plugin after adding it to the project.",
)
@pass_project()
@click.pass_context
def add(
    ctx,
    project: Project,
    plugin_type: str,
    plugin_name: str,
    inherit_from: str = None,
    variant: str = None,
    as_name: str = None,
    **flags,
):
    """
    Add a plugin to your project.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#add
    """
    tracker: Tracker = ctx.obj["tracker"]

    plugin_type = PluginType.from_cli_argument(plugin_type)
    plugin_names = plugin_name  # nargs=-1

    if as_name:
        # `add <type> <inherit-from> --as <name>``
        # is equivalent to:
        # `add <type> <name> --inherit-from <inherit-from>``
        inherit_from = plugin_names[0]
        plugin_names = [as_name]

    plugins_service = ProjectPluginsService(project)

    if flags["custom"]:
        if plugin_type in {
            PluginType.TRANSFORMS,
            PluginType.ORCHESTRATORS,
        }:
            tracker.track_command_event(CliEvent.aborted)
            raise CliError(f"--custom is not supported for {plugin_type}")

    plugin_refs = [
        PluginRef(plugin_type=plugin_type, name=name) for name in plugin_names
    ]
    dependencies_met, err = check_dependencies_met(
        plugin_refs=plugin_refs, plugins_service=plugins_service
    )
    if not dependencies_met:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(f"Failed to install plugin(s): {err}")

    add_service = ProjectAddService(project, plugins_service=plugins_service)

    plugins: list[ProjectPlugin] = []
    for plugin in plugin_names:
        try:
            plugins.append(
                add_plugin(
                    project,
                    plugin_type,
                    plugin,
                    inherit_from=inherit_from,
                    variant=variant,
                    custom=flags["custom"],
                    add_service=add_service,
                )
            )
        except Exception:
            # if the plugin is not known to meltano send what information we do have
            tracker.add_contexts(
                PluginsTrackingContext([(plugin, None) for plugin in plugins])
            )
            tracker.track_command_event(CliEvent.aborted)
            raise

        required_plugins = add_required_plugins(
            project, plugins, add_service=add_service
        )
    plugins.extend(required_plugins)
    tracker.add_contexts(
        PluginsTrackingContext([(candidate, None) for candidate in plugins])
    )
    tracker.track_command_event(CliEvent.inflight)

    if not flags.get("no_install"):
        success = install_plugins(project, plugins, reason=PluginInstallReason.ADD)

        if not success:
            tracker.track_command_event(CliEvent.failed)
            raise CliError("Failed to install plugin(s)")

    _print_plugins(plugins)
    tracker.track_command_event(CliEvent.completed)


def _print_plugins(plugins):
    printed_empty_line = False
    for plugin in plugins:
        docs_url = plugin.docs or plugin.repo
        if not docs_url:
            continue

        if not printed_empty_line:
            click.echo()
            printed_empty_line = True

        click.echo(
            f"To learn more about {plugin.type.descriptor} '{plugin.name}', visit {docs_url}"
        )



"""Manage Python virtual environments."""

from __future__ import annotations

import asyncio
import hashlib
import logging
import os
import platform
import shutil
import subprocess
import sys
from asyncio.subprocess import Process
from collections import namedtuple
from pathlib import Path

from .error import AsyncSubprocessError
from .project import Project

logger = logging.getLogger(__name__)

VenvSpecs = namedtuple("VenvSpecs", ("lib_dir", "bin_dir", "site_packages_dir"))

POSIX = VenvSpecs(
    lib_dir="lib",
    bin_dir="bin",
    site_packages_dir=os.path.join(
        "lib",
        f"python{'.'.join(str(part) for part in sys.version_info[:2])}",
        "site-packages",
    ),
)

NT = VenvSpecs(
    lib_dir="Lib",
    bin_dir="Scripts",
    site_packages_dir=os.path.join("Lib", "site-packages"),
)

PIP_PACKAGES = ("pip", "setuptools==57.5.0", "wheel")


class VirtualEnv:
    PLATFORM_SPECS = {"Linux": POSIX, "Darwin": POSIX, "Windows": NT}

    def __init__(self, root: Path):
        self.root = root
        self._specs = self.__class__.platform_specs()

    @classmethod
    def platform_specs(cls):
        system = platform.system()
        try:
            return cls.PLATFORM_SPECS[system]
        except KeyError:
            raise Exception(f"Platform {system} not supported.")

    def __getattr__(self, attr):
        spec = getattr(self._specs, attr)
        return self.root.joinpath(spec)

    def __str__(self):
        return str(self.root)


async def exec_async(*args, **kwargs) -> Process:
    """
    Run an executable asyncronously.

    :raises AsyncSubprocessError: if the command fails
    """
    run = await asyncio.create_subprocess_exec(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs,
    )
    await run.wait()

    if run.returncode != 0:
        raise AsyncSubprocessError("Command failed", run)

    return run


def fingerprint(pip_urls: list[str]):
    """Return a unique string for the given pip urls."""
    key = " ".join(sorted(pip_urls))
    return hashlib.sha256(bytes(key, "utf-8")).hexdigest()


class VenvService:
    def __init__(self, project: Project, namespace: str = "", name: str = ""):
        """
        Manage isolated virtual environments.

        The methods in this class are not threadsafe.
        """
        self.project = project
        self.namespace = namespace
        self.name = name
        self.venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        self.python_path = self.venv.bin_dir.joinpath("python")
        self.plugin_fingerprint_path = self.venv.root.joinpath(
            ".meltano_plugin_fingerprint"
        )

    async def install(self, *pip_urls: str, clean: bool = False):
        """
        Configure a virtual environment and install the given `pip_urls` packages in it.

        This will try to use an existing virtual environment if one exists unless commanded
        to `clean`.
        """
        pip_urls = [pip_url for arg in pip_urls for pip_url in arg.split(" ")]

        if not clean and self.requires_clean_install(pip_urls):
            logger.debug(
                f"Packages for '{self.namespace}/{self.name}' have changed so performing a clean install."
            )
            clean = True

        self.clean_run_files()

        if clean:
            await self._clean_install(pip_urls)
        else:
            await self._upgrade_install(pip_urls)
        self.write_fingerprint(pip_urls)

    def requires_clean_install(self, pip_urls: list[str]) -> bool:
        """Return `True` if the virtual environment doesn't exist or can't be reused."""
        meltano_pth_path = self.venv.site_packages_dir.joinpath("meltano_venv.pth")
        if meltano_pth_path.exists():
            # clean up deprecated feature
            return True
        existing_fingerprint = self.read_fingerprint()
        if not existing_fingerprint:
            return True
        return existing_fingerprint != fingerprint(pip_urls)

    def clean_run_files(self):
        """Destroy cached configuration files, if they exist."""
        try:
            shutil.rmtree(self.project.run_dir(self.name, make_dirs=False))
        except FileNotFoundError:
            logger.debug("No cached configuration files to remove")

    def clean(self):
        """Destroy the virtual environment, if it exists."""
        try:
            shutil.rmtree(self.venv.root)
            logger.debug(
                "Removed old virtual environment for '%s/%s'",  # noqa: WPS323
                self.namespace,
                self.name,
            )
        except FileNotFoundError:
            # If the VirtualEnv has never been created before do nothing
            logger.debug("No old virtual environment to remove")

    async def create(self):
        """Create a new virtual environment."""
        logger.debug(f"Creating virtual environment for '{self.namespace}/{self.name}'")
        try:
            return await exec_async(
                sys.executable,
                "-m",
                "venv",
                str(self.venv),
            )
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Could not create the virtualenv for '{self.namespace}/{self.name}'",
                err.process,
            )

    async def upgrade_pip(self):
        """Upgrade the `pip` package to the latest version in the virtual environment."""
        logger.debug(f"Upgrading pip for '{self.namespace}/{self.name}'")

        try:
            return await self._pip_install(*PIP_PACKAGES, upgrade=True)

        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                "Failed to upgrade pip to the latest version.", err.process
            )

    def read_fingerprint(self) -> str | None:
        """Return the fingerprint of the existing virtual environment, if any."""
        if not self.plugin_fingerprint_path.exists():
            return None
        with open(self.plugin_fingerprint_path) as fingerprint_file:
            return fingerprint_file.read()

    def write_fingerprint(self, pip_urls: list[str]):
        """Save the fingerprint for this installation."""
        with open(self.plugin_fingerprint_path, "wt") as fingerprint_file:
            fingerprint_file.write(fingerprint(pip_urls))

    def exec_path(self, executable: str) -> Path:
        """Return the absolute path for the given binary in the virtual environment."""
        return self.venv.bin_dir.joinpath(executable)

    async def _clean_install(self, pip_urls: list[str]):
        self.clean()
        await self.create()
        await self.upgrade_pip()

        logger.debug(
            f"Installing '{' '.join(pip_urls)}' into virtual environment for '{self.namespace}/{self.name}'"  # noqa: WPS221
        )
        await self._pip_install(*pip_urls)

    async def _upgrade_install(self, pip_urls: list[str]):
        logger.debug(
            f"Upgrading '{' '.join(pip_urls)}' in existing virtual environment for '{self.namespace}/{self.name}'"  # noqa: WPS221
        )
        await self._pip_install(*PIP_PACKAGES, *pip_urls, upgrade=True)

    async def _pip_install(self, *pip_urls: str, upgrade: bool = False):
        """
        Install a package using `pip` in the proper virtual environment.

        :raises: AsyncSubprocessError: if the command fails.
        """
        args = [
            str(self.python_path),
            "-m",
            "pip",
            "install",
        ]
        if upgrade:
            args += ["--upgrade"]
        args += pip_urls

        try:
            return await exec_async(*args)
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Failed to install plugin '{self.name}'.", err.process
            )



"""Manager for state that persists across runs of a BlockSet.

'State' in this module refers to _Singer_ state, which is held in a Job's
'payload' field. This is not to be confused with the Job's 'state' field,
which refers to a given job run's status, e.g. 'RUNNING' or 'FAILED'.
"""
from __future__ import annotations

import datetime
import json
from collections import defaultdict
from typing import Any

import structlog

from meltano.core.job import Job, JobFinder, Payload, State
from meltano.core.utils import merge

logger = structlog.getLogger(__name__)


class InvalidJobStateError(Exception):
    """Occurs when invalid job state is parsed."""


class StateService:
    """Meltano Service used to manage job state.

    Currently only manages Singer state for Extract and Load jobs.
    """

    def __init__(self, session: object = None):
        """Create a StateService object.

        Args:
            session: the session to use for interacting with the db
        """
        self.session = session

    def list_state(self, state_id_pattern: str | None = None) -> dict[str, dict]:
        """List all state found in the db.

        Args:
            state_id_pattern: An optional glob-style pattern of state_ids to search for

        Returns:
            A dict with state_ids as keys and state payloads as values.
        """
        states = defaultdict(dict)
        query = self.session.query(Job)
        if state_id_pattern:
            query = query.filter(Job.job_name.like(state_id_pattern.replace("*", "%")))
        for state_id in {job.job_name for job in query}:  # noqa: WPS335
            states[state_id] = self.get_state(state_id)
        return states

    def _get_or_create_job(self, job: Job | str) -> Job:
        """If Job is passed, return it. If state_id is passed, create new and return.

        Args:
            job: either an existing Job to modify state for, or a state_id

        Raises:
            TypeError: if job is not of type Job or str

        Returns:
            A new job with given state_id, or the given Job
        """
        if isinstance(job, str):
            now = datetime.datetime.utcnow()
            return Job(
                job_name=job, state=State.STATE_EDIT, started_at=now, ended_at=now
            )
        elif isinstance(job, Job):
            return job
        raise TypeError("job must be of type Job or of type str")

    @staticmethod
    def validate_state(state: dict[str, Any]):
        """Check that the given state str is valid.

        Args:
            state: the state to validate

        Raises:
            InvalidJobStateError: if supplied state is not valid singer state
        """
        if "singer_state" not in state:
            raise InvalidJobStateError(
                "singer_state not found in top level of provided state"
            )

    def add_state(
        self,
        job: Job | str,
        new_state: str | None,
        payload_flags: Payload = Payload.STATE,
        validate=True,
    ):
        """Add state for the given Job.

        Args:
            job: either an existing Job or a state_id that future runs may look up state for.
            new_state: the state to add for the given job.
            payload_flags: the payload_flags to set for the job
            validate: whether to validate the supplied state
        """
        new_state_dict = json.loads(new_state)
        if validate:
            self.validate_state(new_state_dict)
        state_to_add_to = self._get_or_create_job(job)
        state_to_add_to.payload = json.loads(new_state)
        state_to_add_to.payload_flags = payload_flags
        state_to_add_to.save(self.session)
        logger.debug(
            f"Added to state {state_to_add_to.job_name} state payload {new_state_dict}"
        )

    def get_state(self, state_id: str) -> dict:
        """Get state for the given state_id.

        Args:
            state_id: The state_id to get state for

        Returns:
            Dict representing state that would be used in the next run.
        """
        state = {}
        incomplete_since = None
        finder = JobFinder(state_id)

        # Get the state for the most recent completed job.
        # Do not consider dummy jobs create via add_state.
        state_job = finder.latest_with_payload(self.session, flags=Payload.STATE)
        if state_job:
            logger.info(f"Found state from {state_job.started_at}.")
            incomplete_since = state_job.ended_at
            if "singer_state" in state_job.payload:
                merge(state_job.payload, state)

        # If there have been any incomplete jobs since the most recent completed jobs,
        # merge the state emitted by those jobs into the state for the most recent
        # completed job. If there are no completed jobs, get the full history of
        # incomplete jobs and use the most recent state emitted per stream
        incomplete_state_jobs = finder.with_payload(
            self.session, flags=Payload.INCOMPLETE_STATE, since=incomplete_since
        )
        for incomplete_state_job in incomplete_state_jobs:
            logger.info(
                f"Found and merged incomplete state from {incomplete_state_job.started_at}."
            )
            if "singer_state" in incomplete_state_job.payload:
                merge(incomplete_state_job.payload, state)

        return state

    def set_state(self, state_id: str, new_state: str | None, validate: bool = True):
        """Set the state for the state_id.

        Args:
            state_id: the state_id to set state for
            new_state: the state to update to
            validate: whether or not to validate the supplied state.
        """
        self.add_state(
            state_id,
            new_state,
            payload_flags=Payload.STATE,
            validate=validate,
        )

    def clear_state(self, state_id, save: bool = True):
        """Clear the state for the state_id.

        Args:
            state_id: the state_id to clear state for
            save: whether or not to immediately save the state
        """
        self.set_state(state_id, json.dumps({}), validate=False)

    def merge_state(self, state_id_src: str, state_id_dst: str):
        """Merge state from state_id_src into state_id_dst.

        Args:
            state_id_src: the state_id to get state from
            state_id_dst: the state_id_to merge state onto
        """
        src_state_dict = self.get_state(state_id_src)
        src_state = json.dumps(src_state_dict)
        self.add_state(state_id_dst, src_state, payload_flags=Payload.INCOMPLETE_STATE)

    def copy_state(self, state_id_src: str, state_id_dst: str):
        """Copy state from Job state_id_src onto Job state_id_dst.

        Args:
            state_id_src: the state_id to get state from
            state_id_dst: the state_id_to copy state onto
        """
        src_state_dict = self.get_state(state_id_src)
        src_state = json.dumps(src_state_dict)
        self.set_state(state_id_dst, src_state)

    def move_state(self, state_id_src: str, state_id_dst: str):
        """Move state from Job state_id_src to Job state_id_dst.

        Args:
            state_id_src: the state_id to get state from and clear
            state_id_dst: the state_id_to move state onto
        """
        src_state_dict = self.get_state(state_id_src)
        src_state = json.dumps(src_state_dict)
        self.set_state(state_id_dst, src_state)
        self.clear_state(state_id_src)



"""Meltano telemetry."""

from __future__ import annotations

from .contexts import (
    CliContext,
    CliEvent,
    ExceptionContext,
    PluginsTrackingContext,
    ProjectContext,
    environment_context,
)
from .tracker import BlockEvents, Tracker



"""Main entry point for the meltano CLI."""
from __future__ import annotations

import logging
import os
import sys
from typing import TYPE_CHECKING

from meltano.core.logging import setup_logging
from meltano.core.project import ProjectReadonly
from meltano.core.tracking.contexts.exception import ExceptionContext  # noqa: F401

from .utils import CliError

# TODO: Importing the cli.cli module breaks other cli module imports
# This suggests a cyclic dependency or a poorly structured interface.
# This should be investigated and resolved to avoid implicit behavior
# based solely on import order.
from .cli import cli  # isort:skip
from . import (  # isort:skip # noqa: F401, WPS235
    add,
    config,
    discovery,
    dragon,
    elt,
    environment,
    initialize,
    install,
    invoke,
    lock,
    remove,
    repl,
    schedule,
    schema,
    select,
    state,
    ui,
    upgrade,
    user,
    run,
    validate,
    job,
)

if TYPE_CHECKING:
    from meltano.core.tracking.tracker import Tracker


# Holds the exit code for error reporting during process exiting. In particular, a function
# registered by the `atexit` module uses this value.
exit_code: None | int = None

atexit_handler_registered = False
exit_code_reported = False
exit_event_tracker: Tracker = None

setup_logging()

logger = logging.getLogger(__name__)


def _run_cli():
    """Run the Meltano CLI."""
    try:
        try:  # noqa: WPS505
            cli(obj={"project": None})
        except ProjectReadonly as err:
            raise CliError(
                f"The requested action could not be completed: {err}"
            ) from err
        except KeyboardInterrupt:  # noqa: WPS329
            raise
        except Exception as err:
            raise CliError(str(err)) from err
    except CliError as cli_error:
        cli_error.print()
        sys.exit(1)


def main():
    """Entry point for the meltano CLI."""
    # Mark the current process as executed via the CLI
    os.environ["MELTANO_JOB_TRIGGER"] = os.getenv("MELTANO_JOB_TRIGGER", "cli")
    try:
        _run_cli()
    finally:
        global exit_code
        ex = sys.exc_info()[1]
        if ex is None:
            exit_code = 0  # noqa: WPS442
        elif isinstance(ex, SystemExit):
            exit_code = 0 if ex.code is None else ex.code  # noqa: WPS442
        else:
            exit_code = 1  # noqa: WPS442
        # Track the exit event now to provide more details via the exception context.
        # We assume the process will exit practically immediately after `main` returns.
        if exit_event_tracker is not None:
            exit_event_tracker.track_exit_event()



"""API Plugin Management Blue Print."""

from __future__ import annotations

import asyncio
import logging

from flask import jsonify, request

from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.auth import block_if_readonly
from meltano.core.error import PluginInstallError
from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin_install_service import (
    PluginInstallReason,
    PluginInstallService,
)
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_plugins_service import ProjectPluginsService


def plugin_def_json(plugin_def):
    """Convert plugin defenition to json.

    Args:
        plugin_def: Plugin definition

    Returns:
        JSON of the plugin's definition
    """
    return {
        "name": plugin_def.name,
        "namespace": plugin_def.namespace,
        "hidden": plugin_def.hidden,
        "label": plugin_def.label,
        "logo_url": plugin_def.logo_url,
        "description": plugin_def.description,
        "variants": [
            {
                "name": v.name,  # noqa: WPS111
                "default": i == 0,  # noqa: WPS111
                "deprecated": v.deprecated,
            }
            for i, v in enumerate(plugin_def.variants)  # noqa: WPS111
        ],
    }


plugins_bp = APIBlueprint("plugins", __name__)  # noqa: N816


@plugins_bp.errorhandler(PluginInstallError)
def _handle(ex):
    return (jsonify({"error": True, "code": str(ex)}), 502)


@plugins_bp.route("/all", methods=["GET"])  # noqa: WPS125
def all():
    """Plugins found by the PluginDiscoveryService.

    Returns:
        Json containing all the discovered plugins.
    """
    project = Project.find()
    discovery = PluginDiscoveryService(project)

    all_plugins = {
        plugin_type: [plugin_def_json(plugin_def) for plugin_def in plugin_defs]
        for plugin_type, plugin_defs in discovery.plugins_by_type().items()
    }

    return jsonify(all_plugins)


@plugins_bp.route("/installed", methods=["GET"])
def installed():
    """All plugins installed in the project.

    Returns:
        Json of all installed plugins.
    """
    project = Project.find()
    plugins_service = ProjectPluginsService(project)

    def _plugin_json(plugin: ProjectPlugin):
        plugin_json = {"name": plugin.name}

        try:
            plugin_json.update(plugin_def_json(plugin))

            plugin_json["variant"] = plugin.variant
            plugin_json["docs"] = plugin.docs
        except PluginNotFoundError:
            pass

        return plugin_json

    installed_plugins = {
        plugin_type: [_plugin_json(plugin) for plugin in plugins]
        for plugin_type, plugins in plugins_service.plugins_by_type().items()
    }

    return jsonify(installed_plugins)


@plugins_bp.route("/add", methods=["POST"])
@block_if_readonly
def add():
    """Add Plugin the the project file.

    Returns:
        JSON of the plugin information added.
    """
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]
    variant = payload.get("variant", None)

    project = Project.find()
    add_service = ProjectAddService(project)
    plugin = add_service.add(plugin_type, plugin_name, variant=variant)

    return jsonify(plugin.canonical())


@plugins_bp.route("/install/batch", methods=["POST"])
@block_if_readonly
def install_batch():  # noqa: WPS210
    """Install multiple plugins at once.

    Raises:
        PluginInstallError: Plugin insatllation error message.

    Returns:
        JSON cotaining all plugins installed.
    """
    payload = request.get_json()
    project = Project.find()

    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.find_plugin(
        payload["name"], plugin_type=PluginType(payload["plugin_type"])
    )

    add_service = ProjectAddService(project, plugins_service=plugins_service)
    required_plugins = add_service.add_required(plugin)

    # This was added to assist api_worker threads
    try:
        asyncio.get_event_loop()
    except RuntimeError:
        logging.debug("/plugins/install/batch no asyncio event loop detected")
        asyncio.set_event_loop(asyncio.new_event_loop())

    install_service = PluginInstallService(project, plugins_service=plugins_service)
    install_results = install_service.install_plugins(
        required_plugins, reason=PluginInstallReason.ADD
    )

    for result in install_results:
        if not result.successful:
            raise PluginInstallError(result.message)

    return jsonify([plugin.canonical() for plugin in required_plugins])


@plugins_bp.route("/install", methods=["POST"])
@block_if_readonly
def install():
    """Install a plugin.

    Returns:
        JSON containing the plugin installed.
    """
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()

    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.find_plugin(plugin_name, plugin_type=plugin_type)

    # This was added to assist api_worker threads
    try:
        asyncio.get_event_loop()
    except RuntimeError:
        logging.debug("/plugins/install no asyncio event loop detected")
        asyncio.set_event_loop(asyncio.new_event_loop())

    install_service = PluginInstallService(project, plugins_service=plugins_service)
    install_service.install_plugin(plugin, reason=PluginInstallReason.ADD)

    return jsonify(plugin.canonical())



"""Defines `meltano elt` command."""
from __future__ import annotations

import datetime
import logging
import platform
from contextlib import asynccontextmanager, nullcontext

import click
import structlog
from structlog import stdlib as structlog_stdlib

from meltano.core.db import project_engine
from meltano.core.elt_context import ELTContextBuilder
from meltano.core.job import Job, JobFinder
from meltano.core.job.stale_job_failer import fail_stale_jobs
from meltano.core.logging import JobLoggingService, OutputLogger
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.runner import RunnerError
from meltano.core.runner.dbt import DbtRunner
from meltano.core.runner.singer import SingerRunner
from meltano.core.tracking import CliEvent, PluginsTrackingContext, Tracker
from meltano.core.utils import click_run_async

from . import cli
from .params import pass_project
from .utils import CliError, PartialInstrumentedCmd

DUMPABLES = {
    "catalog": (PluginType.EXTRACTORS, "catalog"),
    "state": (PluginType.EXTRACTORS, "state"),
    "extractor-config": (PluginType.EXTRACTORS, "config"),
    "loader-config": (PluginType.LOADERS, "config"),
}

logger = structlog_stdlib.get_logger(__name__)


@cli.command(
    cls=PartialInstrumentedCmd,
    short_help="Run an ELT pipeline to Extract, Load, and Transform data.",
)
@click.argument("extractor")
@click.argument("loader")
@click.option("--transform", type=click.Choice(["skip", "only", "run"]))
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option(
    "--full-refresh",
    help="Perform a full refresh (ignore state left behind by any previous runs).",
    is_flag=True,
)
@click.option(
    "--select",
    "-s",
    help="Select only these specific entities for extraction.",
    multiple=True,
    default=[],
)
@click.option(
    "--exclude",
    "-e",
    help="Exclude these specific entities from extraction.",
    multiple=True,
    default=[],
)
@click.option("--catalog", help="Extractor catalog file.")
@click.option("--state", help="Extractor state file.")
@click.option(
    "--dump",
    type=click.Choice(DUMPABLES.keys()),
    help="Dump content of pipeline-specific generated file.",
)
@click.option(
    "--state-id", envvar="MELTANO_STATE_ID", help="A custom string to identify the job."
)
@click.option(
    "--force",
    "-f",
    help="Force a new run even when a pipeline with the same state ID is already running.",
    is_flag=True,
)
@click.pass_context
@pass_project(migrate=True)
@click_run_async
async def elt(
    project: Project,
    ctx: click.Context,
    extractor: str,
    loader: str,
    transform: str,
    dry: bool,
    full_refresh: bool,
    select: list[str],
    exclude: list[str],
    catalog: str,
    state: str,
    dump: str,
    state_id: str,
    force: bool,
):
    """
    Run an ELT pipeline to Extract, Load, and Transform data.

    meltano elt '<extractor_name>' '<loader_name>'

    extractor_name: extractor to be used in this pipeline.
    loader_name: loader to be used in this pipeline.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#elt
    """
    if platform.system() == "Windows":
        raise CliError(
            "ELT command not supported on Windows. Please use the Run command as documented here https://docs.meltano.com/reference/command-line-interface#run"
        )

    tracker: Tracker = ctx.obj["tracker"]

    # we no longer set a default choice for transform, so that we can detect explicit usages of the --transform option
    # if transform is None we still need manually default to skip after firing the tracking event above.
    if not transform:
        transform = "skip"

    select_filter = [*select, *(f"!{entity}" for entity in exclude)]

    job = Job(
        job_name=state_id
        or f'{datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S")}--{extractor}--{loader}'
    )
    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    try:
        plugins_service = ProjectPluginsService(project)
        context_builder = _elt_context_builder(
            project,
            job,
            session,
            extractor,
            loader,
            transform,
            dry_run=dry,
            full_refresh=full_refresh,
            select_filter=select_filter,
            catalog=catalog,
            state=state,
            plugins_service=plugins_service,
        )

        if dump:
            await dump_file(context_builder, dump)
        else:
            await _run_job(tracker, project, job, session, context_builder, force=force)
    except Exception as err:
        tracker.track_command_event(CliEvent.failed)
        raise err
    finally:
        session.close()

    tracker.track_command_event(CliEvent.completed)


def _elt_context_builder(
    project,
    job,
    session,
    extractor,
    loader,
    transform,
    dry_run=False,
    full_refresh=False,
    select_filter=None,
    catalog=None,
    state=None,
    plugins_service=None,
):
    select_filter = select_filter or []
    transform_name = None
    if transform != "skip":
        transform_name = _find_transform_for_extractor(
            extractor, plugins_service=plugins_service
        )

    return (
        ELTContextBuilder(project, plugins_service=plugins_service)  # noqa: WPS221
        .with_session(session)
        .with_job(job)
        .with_extractor(extractor)
        .with_loader(loader)
        .with_transform(transform_name or transform)
        .with_dry_run(dry_run)
        .with_only_transform(transform == "only")
        .with_full_refresh(full_refresh)
        .with_select_filter(select_filter)
        .with_catalog(catalog)
        .with_state(state)
    )


async def dump_file(context_builder, dumpable):
    """Dump the given dumpable for this pipeline."""
    elt_context = context_builder.context()

    try:
        plugin_type, file_id = DUMPABLES[dumpable]
        invoker = elt_context.invoker_for(plugin_type)

        async with invoker.prepared(elt_context.session):
            content = await invoker.dump(file_id)

        click.echo(content)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {dumpable} file for this pipeline") from err
    except Exception as err:
        raise CliError(f"Could not dump {dumpable}: {err}") from err


async def _run_job(tracker, project, job, session, context_builder, force=False):
    fail_stale_jobs(session, job.job_name)

    if not force:
        existing = JobFinder(job.job_name).latest_running(session)
        if existing:
            raise CliError(
                f"Another '{job.job_name}' pipeline is already running which started at {existing.started_at}. "
                + "To ignore this check use the '--force' option."
            )

    async with job.run(session):
        job_logging_service = JobLoggingService(project)
        log_file = job_logging_service.generate_log_name(job.job_name, job.run_id)

        output_logger = OutputLogger(log_file)
        context_builder.set_base_output_logger(output_logger)

        log = logger.bind(name="meltano", run_id=str(job.run_id), state_id=job.job_name)

        await _run_elt(tracker, log, context_builder, output_logger)


@asynccontextmanager
async def _redirect_output(log, output_logger):

    meltano_stdout = output_logger.out(
        "meltano", log.bind(stdio="stdout", cmd_type="elt")
    )
    meltano_stderr = output_logger.out(
        "meltano", log.bind(stdio="stderr", cmd_type="elt")
    )

    with meltano_stdout.redirect_logging(ignore_errors=(CliError,)):
        async with meltano_stdout.redirect_stdout(), meltano_stderr.redirect_stderr():  # noqa: WPS316
            try:
                yield
            except CliError as err:
                err.print()
                raise


async def _run_elt(
    tracker: Tracker,
    log: structlog.BoundLogger,
    context_builder: ELTContextBuilder,
    output_logger: OutputLogger,
):
    async with _redirect_output(log, output_logger):
        try:
            elt_context = context_builder.context()
            tracker.add_contexts(PluginsTrackingContext.from_elt_context(elt_context))
            tracker.track_command_event(CliEvent.inflight)

            if elt_context.only_transform:
                log.info("Extract & load skipped.")
            else:
                await _run_extract_load(log, elt_context, output_logger)

            if elt_context.transformer:
                await _run_transform(log, elt_context, output_logger)
            else:
                log.info("Transformation skipped.")
        except RunnerError as err:
            raise CliError(
                f"ELT could not be completed: {err}.\n"
                + "For more detailed log messages re-run the command using 'meltano --log-level=debug ...' CLI flag.\n"
                + f"Note that you can also check the generated log file at '{output_logger.file}'.\n"
                + "For more information on debugging and logging: https://docs.meltano.com/reference/command-line-interface#debugging"
            ) from err


async def _run_extract_load(log, elt_context, output_logger, **kwargs):  # noqa: WPS231

    extractor = elt_context.extractor.name
    loader = elt_context.loader.name

    stderr_log = logger.bind(
        run_id=str(elt_context.job.run_id),
        state_id=elt_context.job.job_name,
        stdio="stderr",
    )

    extractor_log = output_logger.out(extractor, stderr_log.bind(cmd_type="extractor"))
    loader_log = output_logger.out(loader, stderr_log.bind(cmd_type="loader"))

    extractor_out_writer_ctxmgr = nullcontext
    loader_out_writer_ctxmgr = nullcontext
    if logger.getEffectiveLevel() == logging.DEBUG:
        stdout_log = logger.bind(
            run_id=str(elt_context.job.run_id),
            state_id=elt_context.job.job_name,
            stdio="stdout",
        )
        extractor_out = output_logger.out(
            f"{extractor} (out)", stdout_log.bind(cmd_type="extractor"), logging.DEBUG
        )
        loader_out = output_logger.out(
            f"{loader} (out)", stdout_log.bind(cmd_type="loader"), logging.DEBUG
        )

        extractor_out_writer_ctxmgr = extractor_out.line_writer
        loader_out_writer_ctxmgr = loader_out.line_writer

    log.info(
        "Running extract & load...",
    )

    singer_runner = SingerRunner(elt_context)
    try:
        with extractor_log.line_writer() as extractor_log_writer, loader_log.line_writer() as loader_log_writer:
            with extractor_out_writer_ctxmgr() as extractor_out_writer, loader_out_writer_ctxmgr() as loader_out_writer:
                await singer_runner.run(
                    **kwargs,
                    extractor_log=extractor_log_writer,
                    loader_log=loader_log_writer,
                    extractor_out=extractor_out_writer,
                    loader_out=loader_out_writer,
                )
    except RunnerError as err:
        try:  # noqa: WPS505
            code = err.exitcodes[PluginType.EXTRACTORS]
            message = extractor_log.last_line.rstrip() or "(see above)"
            log.error("Extraction failed", code=code, message=message)
        except KeyError:
            pass

        try:  # noqa: WPS505
            code = err.exitcodes[PluginType.LOADERS]
            message = loader_log.last_line.rstrip() or "(see above)"
            log.error("Loading failed", code=code, message=message)
        except KeyError:
            pass

        raise

    log.info("Extract & load complete!")


async def _run_transform(log, elt_context, output_logger, **kwargs):

    stderr_log = logger.bind(
        run_id=str(elt_context.job.run_id),
        state_id=elt_context.job.job_name,
        stdio="stderr",
        cmd_type="transformer",
    )

    transformer_log = output_logger.out(elt_context.transformer.name, stderr_log)

    log.info("Running transformation...")

    dbt_runner = DbtRunner(elt_context)
    try:
        with transformer_log.line_writer() as transformer_log_writer:
            await dbt_runner.run(**kwargs, log=transformer_log_writer)
    except RunnerError as err:
        try:  # noqa: WPS505
            code = err.exitcodes[PluginType.TRANSFORMERS]
            message = transformer_log.last_line.rstrip() or "(see above)"
            log.error("Transformation failed", code=code, message=message)
        except KeyError:
            pass

        raise

    log.info("Transformation complete!")


def _find_transform_for_extractor(extractor: str, plugins_service):
    discovery_service = plugins_service.discovery_service
    try:
        extractor_plugin_def = discovery_service.find_definition(
            PluginType.EXTRACTORS, extractor
        )

        # Check if there is a default transform for this extractor
        transform_plugin_def = discovery_service.find_definition_by_namespace(
            PluginType.TRANSFORMS, extractor_plugin_def.namespace
        )

        # Check if the transform has been added to the project
        transform_plugin = plugins_service.get_plugin(transform_plugin_def)

        return transform_plugin.name
    except PluginNotFoundError:
        return None



"""New Project Initialization Service."""
from __future__ import annotations

import os
import uuid

import click

from .cli_messages import GREETING
from .db import project_engine
from .plugin.meltano_file import MeltanoFilePlugin
from .project import Project
from .project_settings_service import ProjectSettingsService, SettingValueStore


class ProjectInitServiceError(Exception):
    """Project Initialization Service Exception."""


class ProjectInitService:
    """New Project Initialization Service."""

    def __init__(self, project_name):
        """Create a new ProjectInitService instance.

        Args:
            project_name: The name of the project to create
        """
        self.project_name = project_name.lower()

    def init(self, activate: bool = True, add_discovery: bool = False) -> Project:
        """Initialise Meltano Project.

        Args:
            activate: Activate newly created project
            add_discovery: Add discovery.yml file to created project

        Returns:
            A new Project instance

        Raises:
            ProjectInitServiceError: Directory already exists
        """
        try:
            os.mkdir(self.project_name)
        except Exception as exp:  # noqa: F841
            raise ProjectInitServiceError(
                f"Directory {self.project_name} already exists."
            )
        click.secho("Created", fg="blue", nl=False)
        click.echo(f" {self.project_name}")

        self.project = Project(self.project_name)

        self.create_files(add_discovery=add_discovery)

        self.settings_service = ProjectSettingsService(self.project)
        self.settings_service.set(
            "project_id",
            f"{self.project_name}-{uuid.uuid4()}",
            store=SettingValueStore.MELTANO_YML,
        )
        self.set_send_anonymous_usage_stats()
        if activate:
            Project.activate(self.project)

        self.create_system_database()

        return self.project

    def create_dot_meltano_dir(self):
        """Create .meltano directory."""
        # explicitly create the .meltano directory if it doesn't exist
        os.makedirs(self.project.meltano_dir(), exist_ok=True)
        click.secho("   |--", fg="blue", nl=False)
        click.echo(f" {self.project.meltano_dir().name}")

    def create_files(self, add_discovery=False):
        """Create project files.

        Args:
            add_discovery: Add discovery.yml file to created project
        """
        click.secho("Creating project files...", fg="blue")
        click.echo(f"  {self.project_name}/")

        self.create_dot_meltano_dir()

        plugin = MeltanoFilePlugin(discovery=add_discovery)
        for path in plugin.create_files(self.project):
            click.secho("   |--", fg="blue", nl=False)
            click.echo(f" {path}")

    def set_send_anonymous_usage_stats(self):
        """Set Anonymous Usage Stats flag."""
        # If set to false store explicitly in `meltano.yml`
        if not self.settings_service.get("send_anonymous_usage_stats"):
            self.settings_service.set(
                "send_anonymous_usage_stats",
                self.settings_service.get("send_anonymous_usage_stats"),
                store=SettingValueStore.MELTANO_YML,
            )

    def create_system_database(self):
        """Create Meltano System DB.

        Raises:
            ProjectInitServiceError: Database initialization failed
        """
        click.secho("Creating system database...", fg="blue", nl=False)

        # register the system database connection
        engine, _ = project_engine(self.project, default=True)

        from meltano.core.migration_service import MigrationError, MigrationService

        try:
            migration_service = MigrationService(engine)
            migration_service.upgrade(silent=True)
            migration_service.seed(self.project)
            click.secho("  Done!", fg="blue")
        except MigrationError as err:
            raise ProjectInitServiceError(str(err)) from err

    def echo_instructions(self):
        """Echo Next Steps to Click CLI."""
        click.secho(GREETING, nl=False)
        click.secho("\nProject ", nl=False)
        click.secho(self.project_name, fg="magenta", nl=False)
        click.echo(" has been created!\n")

        click.echo("Meltano Environments initialized with ", nl=False)
        click.secho("dev", fg="bright_green", nl=False)
        click.echo(", ", nl=False)
        click.secho("staging", fg="bright_yellow", nl=False)
        click.echo(", and ", nl=False)
        click.secho("prod", fg="bright_red", nl=False)
        click.echo(".")
        click.echo("To learn more about Environments visit: ", nl=False)
        click.secho(
            "https://docs.meltano.com/concepts/environments",
            fg="cyan",
        )

        click.echo("\nNext steps:")
        click.secho("  cd ", nl=False)
        click.secho(self.project_name, fg="magenta")
        click.echo("  Visit ", nl=False)
        click.secho(
            "https://docs.meltano.com/getting-started#create-your-meltano-project",
            fg="cyan",
            nl=False,
        )
        click.echo(" to learn where to go from here")

    def join_with_project_base(self, filename):
        """Join Path to Project base.

        Args:
            filename: File name to join with project base

        Returns:
            Joined base path and passed filename
        """
        return os.path.join(".", self.project_name, filename)



"""Meltano Iglu schemas metadata & utilities."""

from __future__ import annotations

from dataclasses import dataclass

DEFAULT_VENDOR = "com.meltano"


@dataclass
class IgluSchema:
    """Dataclass to store the name, version, vendor, and URL for an Iglu schema."""

    name: str
    version: str
    vendor: str = DEFAULT_VENDOR

    @property
    def url(self) -> str:
        """Construct an iglu schema URL.

        Returns:
            The URL to the schema.
        """
        return f"iglu:{self.vendor}/{self.name}/jsonschema/{self.version}"


CliContextSchema = IgluSchema("cli_context", "1-1-0")
CliEventSchema = IgluSchema("cli_event", "1-0-1")
BlockEventSchema = IgluSchema("block_event", "1-0-0")
EnvironmentContextSchema = IgluSchema("environment_context", "1-0-0")
ExceptionContextSchema = IgluSchema("exception_context", "1-0-0")
ExitEventSchema = IgluSchema("exit_event", "1-0-0")
PluginsContextSchema = IgluSchema("plugins_context", "1-0-0")
ProjectContextSchema = IgluSchema("project_context", "1-1-0")
TelemetryStateChangeEventSchema = IgluSchema("telemetry_state_change_event", "1-0-0")



"""CLI command `meltano install`."""

from __future__ import annotations

import click

from meltano.cli import cli
from meltano.cli.params import pass_project
from meltano.cli.utils import CliError, PartialInstrumentedCmd, install_plugins
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.tracking import CliEvent, PluginsTrackingContext, Tracker


@cli.command(cls=PartialInstrumentedCmd, short_help="Install project dependencies.")
@click.argument(
    "plugin_type", type=click.Choice(PluginType.cli_arguments()), required=False
)
@click.argument("plugin_name", nargs=-1, required=False)
@click.option(
    "--clean",
    is_flag=True,
    help="Completely reinstall a plugin rather than simply upgrading if necessary.",
)
@click.option(
    "--parallelism",
    "-p",
    type=click.INT,
    default=None,
    help="Limit the number of plugins to install in parallel. Defaults to the number of cores.",
)
@click.pass_context
@pass_project(migrate=True)
def install(
    project: Project,
    ctx: click.Context,
    plugin_type: str,
    plugin_name: str,
    clean: bool,
    parallelism: int,
):
    """
    Install all the dependencies of your project based on the meltano.yml file.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#install
    """
    tracker: Tracker = ctx.obj["tracker"]

    plugins_service = ProjectPluginsService(project)

    try:
        if plugin_type:
            plugin_type = PluginType.from_cli_argument(plugin_type)
            plugins = plugins_service.get_plugins_of_type(plugin_type)
            if plugin_name:
                plugins = [plugin for plugin in plugins if plugin.name in plugin_name]
        else:
            plugins = list(plugins_service.plugins())
    except Exception:
        tracker.track_command_event(CliEvent.aborted)
        raise

    click.echo(f"Installing {len(plugins)} plugins...")
    tracker.add_contexts(
        PluginsTrackingContext([(candidate, None) for candidate in plugins])
    )
    tracker.track_command_event(CliEvent.inflight)

    success = install_plugins(project, plugins, parallelism=parallelism, clean=clean)
    if not success:
        tracker.track_command_event(CliEvent.failed)
        raise CliError("Failed to install plugin(s)")
    tracker.track_command_event(CliEvent.completed)



"""Migration and system db management."""
from __future__ import annotations

import logging
from contextlib import closing

import click
import sqlalchemy
from alembic import command
from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

from meltano.api.models.security import Role, RolePermissions
from meltano.core.db import project_engine
from meltano.core.project import Project
from meltano.migrations import LOCK_PATH, MIGRATION_DIR

SPLAT = "*"


class MigrationError(Exception):
    """Generic class for migration errors."""


class MigrationUneededException(Exception):
    """Occurs when no migrations are needed."""


class MigrationService:
    """Migration service."""

    def __init__(self, engine: Engine) -> None:
        """Initialize the migration service.

        Args:
            engine: The sqlalchemy engine to use for the migration and checks.
        """
        self.engine = engine

    def ensure_migration_needed(
        self, script: ScriptDirectory, context: MigrationContext, target_revision: str
    ) -> None:
        """Ensure that a migration of the system database is actually needed.

        Args:
            script: The alembic script directory.
            context: The migration context.
            target_revision: The desired target revision.

        Raises:
            MigrationUneededException: If no migration is needed.
        """
        current_head = context.get_current_revision()

        for rev in script.iterate_revisions(current_head, "base"):
            if rev.revision == target_revision:
                raise MigrationUneededException

    def upgrade(  # noqa: WPS213, WPS231 too many expression and too complex
        self, silent: bool = False
    ) -> None:
        """Upgrade to the latest revision.

        Args:
            silent: If true, don't print anything.

        Raises:
            MigrationError: If the upgrade fails.
        """
        conn = self.engine.connect()
        cfg = Config()

        # this connection is used in `env.py` for the migrations
        cfg.attributes["connection"] = conn
        cfg.set_main_option("script_location", str(MIGRATION_DIR))
        script = ScriptDirectory.from_config(cfg)
        # let's make sure we actually need to migrate

        migration_logger = logging.getLogger("alembic.runtime.migration")
        original_log_level = migration_logger.getEffectiveLevel()
        if silent:
            migration_logger.setLevel(logging.ERROR)

        context = MigrationContext.configure(conn)

        try:  # noqa: WPS229
            # try to find the locked version
            head = LOCK_PATH.open().read().strip()
            self.ensure_migration_needed(script, context, head)

            if not silent:
                click.secho(f"Upgrading database to {head}")
            command.upgrade(cfg, head)

            if silent:
                migration_logger.setLevel(original_log_level)
        except FileNotFoundError:
            raise MigrationError(
                "Cannot upgrade the system database, revision lock not found."
            )
        except MigrationUneededException:
            if not silent:
                click.secho("System database up-to-date.")
        except Exception as err:
            logging.exception(str(err))
            raise MigrationError(
                "Cannot upgrade the system database. It might be corrupted or was created before database migrations where introduced (v0.34.0)"
            )
        finally:
            conn.close()

    def seed(self, project: Project) -> None:
        """Seed the database with the default roles and permissions.

        Args:
            project: The project to seed the database for.
        """
        _, session_maker = project_engine(project)
        with closing(session_maker()) as session:
            self._create_user_role(session)
            session.commit()

    def _create_user_role(self, session: Session) -> None:
        """Actually perform the database seeding creating users/roles.

        Args:
            session: The session to use.
        """
        if not session.query(Role).filter_by(name="admin").first():

            session.add(
                Role(
                    name="admin",
                    description="Meltano Admin",
                    permissions=[
                        RolePermissions(type="view:design", context=SPLAT),
                        RolePermissions(type="view:reports", context=SPLAT),
                        RolePermissions(type="modify:acl", context=SPLAT),
                    ],
                )
            )

        if not session.query(Role).filter_by(name="regular").first():
            session.merge(Role(name="regular", description="Meltano User"))

        # add the universal permissions to Admin
        admin = session.query(Role).filter_by(name="admin").one()
        try:
            session.query(RolePermissions).filter_by(
                role=admin, type=SPLAT, context=SPLAT
            ).one()
        except sqlalchemy.orm.exc.NoResultFound:
            admin.permissions.append(RolePermissions(type=SPLAT, context=SPLAT))



from __future__ import annotations

from . import PluginRef


class PluginNotFoundError(Exception):
    """Base exception when a plugin could not be found."""

    def __init__(self, plugin_or_name):
        if isinstance(plugin_or_name, PluginRef):
            self.plugin_type = plugin_or_name.type.descriptor
            self.plugin_name = plugin_or_name.name
        else:
            self.plugin_type = "plugin"
            self.plugin_name = plugin_or_name

    def __str__(self):
        return f"{self.plugin_type.capitalize()} '{self.plugin_name}' is not known to Meltano"


class PluginParentNotFoundError(Exception):
    """Base exception when a plugin's parent could not be found."""

    def __init__(self, plugin, parent_not_found_error):
        """Initialize exception for when plugin's parent could not be found."""
        self.plugin = plugin
        self.parent_not_found_error = parent_not_found_error

    def __str__(self):
        return "Could not find parent plugin for {type} '{name}': {error}".format(
            type=self.plugin.type.descriptor,
            name=self.plugin.name,
            error=self.parent_not_found_error,
        )


class PluginNotSupportedError(Exception):
    """Base exception when a plugin is not supported for some operation."""

    def __init__(self, plugin: PluginRef):
        """Construct a PluginNotSupportedError instance."""
        self.plugin = plugin

    def __str__(self):
        return f"Operation not supported for {self.plugin.type.descriptor} '{self.plugin.name}'"


class PluginExecutionError(Exception):
    """Base exception for problems that stem from the execution of a plugin (sub-process)."""


class PluginLacksCapabilityError(Exception):
    """Base exception when a plugin lacks a requested capability."""



"""add_state_edit_to_job_state_enum

Revision ID: 13e8639c6d2b
Revises: d135f52a6f49
Create Date: 2022-04-21 09:35:35.435614

"""
from enum import Enum

import sqlalchemy as sa
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "13e8639c6d2b"
down_revision = "d135f52a6f49"
branch_labels = None
depends_on = None


# from core/job/job.py
class State(Enum):
    """Represents status of a Job."""

    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())
    STATE_EDIT = (5, ())


def upgrade():
    dialect_name = get_dialect_name()
    max_string_length = max_string_length_for_dialect(dialect_name)

    conn = op.get_bind()
    # In sqlite, the field is already a varchar.
    # "ALTER COLUMN" statements are also not supported.
    if conn.dialect.name != "sqlite":
        op.alter_column(
            table_name="job",
            column_name="state",
            type_=sa.types.String(max_string_length),
            existing_type=sa.Enum(State, name="job_state"),
            existing_nullable=True,
        )

    # In postgresql, drop the created Enum type so that
    # downgrade() can re-create it.
    if conn.dialect.name == "postgresql":
        conn.execute("DROP TYPE job_state;")


def downgrade():
    conn = op.get_bind()
    # In sqlite, the field is already a varchar.
    # "ALTER COLUMN" statements are also not supported.
    if conn.dialect.name != "sqlite":
        op.alter_column(
            table_name="job",
            column_name="state",
            _type=sa.Enum(State, name="job_state"),
            existing_type=sa.types.String,
        )



from __future__ import annotations

from copy import copy
from typing import Callable

from flask_principal import Need

from .auth import ResourcePermission


class ResourceFilter:
    """Filters API responses for a certain set of permissions."""

    def __init__(self):
        self._needs = []

    def filter_all(self, permission_type, resources):
        return list(self.filter(permission_type, resources))

    def filter(self, permission_type, resources):
        for r in resources:
            needs = [need(permission_type, r) for need in self._needs]
            perm = ResourcePermission(*needs)

            if perm.can():
                yield self.scope(permission_type, copy(r))

    def needs(self, need: Callable):
        self._needs.append(need)

        return self

    def scope(self, permission_type, resource):
        return resource


class NameFilterMixin:
    def __init__(self, *args):
        super().__init__(*args)

        self.needs(self.name_need)

    def name_need(self, permission_type, resource):
        return Need(permission_type, resource["name"])


class TopicFilter(NameFilterMixin, ResourceFilter):
    def __init__(self, *args, design_filter=None):
        super().__init__(*args)

        self._design_filter = design_filter or DesignFilter()

    def scope(self, permission_type, topic):
        topic["designs"] = list(
            self._design_filter.filter("view:design", topic["designs"])
        )

        return topic


class DesignFilter(NameFilterMixin, ResourceFilter):
    pass  # noqa: WPS604



"""The block module implements and supports the Meltano 'block' architecture.

Currently comprised of:

BlockSet interface spec
- ExtractLoadBlock - a block set implementing basic ELT functionality using lower level IOBlocks.
IOBlock interface spec
- SingerBlock - a IOBlock implementation wrapping singer plugins.
"""



"""Discover plugin definitions."""

from __future__ import annotations

import io
import logging
import re
from abc import ABCMeta, abstractmethod
from typing import Iterable

import requests
from ruamel.yaml import YAMLError

import meltano
from meltano.core import bundle
from meltano.core.behavior.versioned import IncompatibleVersionError, Versioned
from meltano.core.discovery_file import DiscoveryFile
from meltano.core.plugin import BasePlugin, PluginDefinition, PluginRef, PluginType
from meltano.core.plugin.base import StandalonePlugin
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.factory import base_plugin_factory
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import NotFound, find_named
from meltano.core.yaml import yaml


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""


# Increment this version number whenever the schema of discovery.yml is changed.
# See https://docs.meltano.com/contribute/plugins#discoveryyml-version for more information.
VERSION = 22


class PluginRepository(metaclass=ABCMeta):
    """A generic plugin definition repository."""

    @abstractmethod
    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        **kwargs,
    ) -> PluginDefinition:
        """Find a plugin definition.

        Args:
            plugin_type: The type of plugin to find.
            plugin_name: The name of the plugin to find.
            kwargs: Additional arguments to pass to the finder.
        """
        ...  # noqa: WPS428

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The type of plugin to get the base plugin for.
            plugin_name: The name of the plugin to get the base plugin for.
            variant: The variant of the plugin to get the base plugin for.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
        )

        return base_plugin_factory(plugin, variant)

    def get_base_plugin(
        self,
        project_plugin: ProjectPlugin,
        **kwargs,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            project_plugin: The project plugin to get the base plugin for.
            kwargs: Additional arguments to pass to the finder.

        Returns:
            The base plugin.
        """
        plugin = project_plugin.custom_definition or self.find_definition(
            project_plugin.type,
            project_plugin.inherit_from or project_plugin.name,
            **kwargs,
        )

        return base_plugin_factory(plugin, project_plugin.variant)


class PluginDiscoveryService(  # noqa: WPS214 (too many public methods)
    PluginRepository, Versioned
):
    """Discover plugin definitions."""

    __version__ = VERSION

    def __init__(self, project, discovery: dict | None = None):
        """Create a new PluginDiscoveryService.

        Args:
            project: The project to discover plugins for.
            discovery: The discovery file to use.
        """
        self.project = project

        self._discovery_version = None
        self._discovery = None
        if discovery:
            self._discovery_version = DiscoveryFile.file_version(discovery)
            self._discovery = DiscoveryFile.parse(discovery)

        self.settings_service = ProjectSettingsService(self.project)

    @property
    def file_version(self):
        """Return the version of the discovery file.

        Returns:
            The version of the discovery file.
        """
        return self._discovery_version

    @property
    def discovery_url(self):
        """Return the URL of the discovery file.

        Returns:
            The URL of the discovery file.
        """
        discovery_url = self.settings_service.get("discovery_url")

        if not discovery_url or not re.match(
            r"^https?://", discovery_url  # noqa: WPS360
        ):
            return None

        return discovery_url

    @property
    def discovery_url_auth(self):
        """Return the `discovery_url_auth` setting.

        Returns:
            The `discovery_url_auth` setting.
        """
        return self.settings_service.get("discovery_url_auth")

    @property
    def discovery(self):
        """Return first compatible discovery manifest from a few locations.

        Locations:

        - project local `discovery.yml`
        - `discovery_url` project setting
        - .meltano/cache/discovery.yml
        - meltano.core.bundle

        Returns:
            The discovery file.

        Raises:
            DiscoveryInvalidError: If the discovery file is invalid.
        """
        if self._discovery:
            return self._discovery

        loaders = [
            (
                self.load_local_discovery,
                "your project's local `discovery.yml` manifest",
            ),
            (
                self.load_remote_discovery,
                f"the `discovery.yml` manifest received from {self.discovery_url}",
            ),
            (self.load_cached_discovery, "the cached `discovery.yml` manifest"),
            (self.load_bundled_discovery, "the bundled `discovery.yml` manifest"),
        ]

        errored = False
        for loader, description in loaders:
            if errored:
                logging.warning(f"Falling back on {description}...")

            try:
                loader()
            except IncompatibleVersionError as err:
                errored = True

                logging.warning(
                    f"{description.capitalize()} has version {err.file_version}, while this version of Meltano requires version {err.version}."
                )
                if err.file_version > err.version:
                    logging.warning(
                        "Please install the latest compatible version of Meltano using `meltano upgrade`."
                    )
            except DiscoveryInvalidError as err:
                errored = True
                logging.error(f"{description.capitalize()} could not be parsed.")
                logging.debug(str(err))

            if self._discovery:
                return self._discovery

        raise DiscoveryInvalidError("No valid `discovery.yml` manifest could be found")

    def load_local_discovery(self):
        """Load the local `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        try:
            with self.project.root_dir("discovery.yml").open() as local_discovery:
                return self.load_discovery(local_discovery)
        except FileNotFoundError:
            pass

    def load_remote_discovery(self) -> DiscoveryFile | None:
        """Load the remote `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        discovery_url = self.discovery_url
        if not discovery_url:
            return None

        headers = {"User-Agent": f"Meltano/{meltano.__version__}"}  # noqa: WPS609
        params = {}

        if self.discovery_url_auth:
            headers["Authorization"] = self.discovery_url_auth

        if self.settings_service.get("send_anonymous_usage_stats"):
            project_id = self.settings_service.get("project_id")

            headers["X-Project-ID"] = project_id
            params["project_id"] = project_id

        try:
            response = requests.get(discovery_url, headers=headers, params=params)
            response.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as err:
            logging.debug("Remote `discovery.yml` manifest could not be downloaded.")
            logging.debug(str(err))
            return None

        remote_discovery = io.StringIO(response.text)

        return self.load_discovery(remote_discovery, cache=True)

    def load_cached_discovery(self):
        """Load the cached `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        try:
            with self.cached_discovery_file.open() as cached_discovery:
                return self.load_discovery(cached_discovery)
        except FileNotFoundError:
            pass

    def load_bundled_discovery(self):
        """Load the bundled `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        with open(bundle.root / "discovery.yml") as bundled_discovery:
            discovery = self.load_discovery(bundled_discovery, cache=True)

        return discovery

    def load_discovery(self, discovery_file, cache=False) -> DiscoveryFile:
        """Load the `discovery.yml` manifest.

        Args:
            discovery_file: The file to load.
            cache: Whether to cache the manifest.

        Returns:
            The discovery file.

        Raises:
            DiscoveryInvalidError: If the discovery file is invalid.
        """
        try:
            discovery_yaml = yaml.load(discovery_file)

            self._discovery_version = DiscoveryFile.file_version(discovery_yaml)
            self.ensure_compatible()

            self._discovery = DiscoveryFile.parse(discovery_yaml)

            if cache:
                self.cache_discovery()

            return self._discovery
        except (YAMLError, Exception) as err:
            raise DiscoveryInvalidError(str(err))

    def cache_discovery(self):
        """Cache the `discovery.yml` manifest."""
        with self.cached_discovery_file.open("w") as cached_discovery:
            yaml.dump(
                self._discovery,
                cached_discovery,
            )

    @property
    def cached_discovery_file(self):
        """Return the cached `discovery.yml` manifest file.

        Returns:
            The cached `discovery.yml` manifest file.
        """
        return self.project.meltano_dir("cache", "discovery.yml")

    def get_plugins_of_type(self, plugin_type):
        """Return the plugins of the given type.

        Args:
            plugin_type: The plugin type.

        Returns:
            The list of plugins of the given type.
        """
        return self.discovery[plugin_type]

    def plugins_by_type(self):
        """Return a mapping of plugins by type.

        Returns:
            The plugins by type.
        """
        return {
            plugin_type: self.get_plugins_of_type(plugin_type)
            for plugin_type in PluginType
        }

    def plugins(self) -> Iterable[PluginDefinition]:
        """Generate all plugins.

        Yields:
            Each discoverable plugin.
        """
        yield from (
            plugin
            for plugin_type, plugins in self.plugins_by_type().items()
            for plugin in plugins
        )

    def find_definition(
        self, plugin_type: PluginType, plugin_name: str
    ) -> PluginDefinition:
        """Find a plugin definition by type and name.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin could not be found.
        """
        try:
            return find_named(self.get_plugins_of_type(plugin_type), plugin_name)
        except NotFound as err:
            raise PluginNotFoundError(PluginRef(plugin_type, plugin_name)) from err

    def find_definition_by_namespace(
        self, plugin_type: PluginType, namespace: str
    ) -> PluginDefinition:
        """Find a plugin definition by type and namespace.

        Args:
            plugin_type: The plugin type.
            namespace: The plugin namespace.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin could not be found.
        """
        try:
            return next(
                plugin
                for plugin in self.get_plugins_of_type(plugin_type)
                if plugin.namespace == namespace
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def find_related_plugin_refs(
        self,
        target_plugin: ProjectPlugin,
        plugin_types: list[PluginType] | None = None,
    ):
        """Find related plugin references.

        Args:
            target_plugin: The target plugin.
            plugin_types: Types of related plugins to add.

        Returns:
            The related plugin references.
        """
        plugin_types = plugin_types or list(PluginType)

        try:
            plugin_types.remove(target_plugin.type)
        except ValueError:
            pass

        related_plugin_refs = []

        related_plugin_refs.extend(
            related_plugin_def
            for plugin_type in plugin_types
            for related_plugin_def in self.get_plugins_of_type(plugin_type)
            if related_plugin_def.namespace == target_plugin.namespace
        )

        return related_plugin_refs


class LockedDefinitionService(PluginRepository):
    """PluginRepository implementation for local files."""

    def __init__(self, project: Project) -> None:
        """Initialize the service.

        Args:
            project: The Meltano project.
        """
        self.project = project

    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> PluginDefinition:
        """Find a locked plugin definition.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin definition could not be found.
        """
        path = self.project.plugin_lock_path(plugin_type, plugin_name, variant_name)
        try:
            standalone = StandalonePlugin.parse_json_file(path)
        except FileNotFoundError as err:
            raise PluginNotFoundError(PluginRef(plugin_type, plugin_name)) from err
        return PluginDefinition.from_standalone(standalone)

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant: The plugin variant.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
            variant_name=variant,
        )

        return base_plugin_factory(plugin, plugin.variants[0])



from __future__ import annotations

import logging
import threading
import time
import traceback
import webbrowser

import click
import requests

from meltano.core.project_settings_service import ProjectSettingsService

logger = logging.getLogger(__name__)

SUCCESS_STATUS_CODE = 200


class UIAvailableWorker(threading.Thread):
    def __init__(self, project, open_browser=False):
        super().__init__()
        self.project = project
        self.open_browser = open_browser
        self.settings_service = ProjectSettingsService(self.project)
        self._terminate = False

    def run(self):
        url = f"http://localhost:{self.settings_service.get('ui.bind_port')}"
        headers = {"Host": self.settings_service.get("ui.server_name")}

        while not self._terminate:
            try:
                response = requests.get(url, headers=headers)
                if response.status_code == SUCCESS_STATUS_CODE:
                    click.secho(f"Meltano UI is now available at {url}", fg="green")
                    if self.open_browser:
                        webbrowser.open(url)
                    self._terminate = True
            except Exception:
                logger.debug(
                    f"Exception encountered while trying to run Meltano UI:\n{traceback.format_exc()}"
                )

            time.sleep(2)

    def stop(self):
        self._terminate = True



from __future__ import annotations

import importlib
import logging

from gunicorn.glogging import CONFIG_DEFAULTS

import meltano
from meltano.core.logging.utils import FORMAT
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService

_project = Project.find()
_settings_service = ProjectSettingsService(_project)


loglevel = _settings_service.get("cli.log_level")

logconfig_dict = CONFIG_DEFAULTS.copy()
logconfig_dict["loggers"]["gunicorn.access"]["propagate"] = False
logconfig_dict["loggers"]["gunicorn.error"]["propagate"] = False
logconfig_dict["formatters"]["generic"] = {"format": FORMAT}

if loglevel != "debug":
    logconfig_dict["loggers"].pop("gunicorn.access")

_bind_host = _settings_service.get("ui.bind_host")
_bind_port = _settings_service.get("ui.bind_port")
bind = f"{_bind_host}:{_bind_port}"

workers = _settings_service.get("ui.workers")
forwarded_allow_ips = _settings_service.get("ui.forwarded_allow_ips")

timeout = 600


def on_reload(arbiter):
    importlib.reload(meltano)
    logging.info(f"Meltano version reloaded to {meltano.__version__}")



"""Project Upgrade CLI."""

from __future__ import annotations

import os

import click

from meltano.core.db import project_engine
from meltano.core.meltano_invoker import MeltanoInvoker
from meltano.core.upgrade_service import UpgradeService

from . import cli
from .params import pass_project
from .utils import InstrumentedCmd, InstrumentedDefaultGroup


@cli.group(
    cls=InstrumentedDefaultGroup,
    default="all",
    default_if_no_args=True,
    short_help="Upgrade Meltano and your entire project to the latest version.",
)
@pass_project()
@click.pass_context
def upgrade(ctx, project):
    """
    Upgrade Meltano and your entire project to the latest version.

    When called without arguments, this will:

    - Upgrade the meltano package\n
    - Update files managed by file bundles\n
    - Apply migrations to system database\n

    Read more at https://docs.meltano.com/reference/command-line-interface#upgrade
    """
    engine, _ = project_engine(project)
    upgrade_service = UpgradeService(engine, project)
    ctx.obj["upgrade_service"] = upgrade_service


@upgrade.command(  # noqa: WPS125
    cls=InstrumentedCmd,
    short_help="Upgrade Meltano and your entire project to the latest version.",
)
@click.option(
    "--pip_url", type=str, envvar="MELTANO_UPGRADE_PIP_URL", help="Meltano pip URL."
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    envvar="MELTANO_UPGRADE_FORCE",
    help="Force upgrade.",
)
@click.option(
    "--skip-package",
    is_flag=True,
    default=False,
    help="Skip updating the Meltano package.",
)
@click.pass_context
def all(ctx, pip_url, force, skip_package):
    """
    Upgrade Meltano and your entire project to the latest version.

    When called without arguments, this will:

    - Upgrade the meltano package\n
    - Update files managed by file bundles\n
    - Apply migrations to system database\n

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#upgrade
    """
    project = ctx.obj["project"]
    upgrade_service = ctx.obj["upgrade_service"]

    if skip_package:
        upgrade_service.update_files()

        click.echo()
        upgrade_service.migrate_database()

        click.echo()

        if not os.getenv("MELTANO_PACKAGE_UPGRADED", default=False):
            click.echo()
            click.secho("Your Meltano project has been upgraded!", fg="green")
    else:
        package_upgraded = upgrade_service.upgrade_package(pip_url=pip_url, force=force)
        if package_upgraded:
            # Shell out instead of calling `upgrade_service` methods to
            # ensure the latest code is used.
            click.echo()
            run = MeltanoInvoker(project).invoke(
                ["upgrade", "--skip-package"],
                env={"MELTANO_PACKAGE_UPGRADED": "true"},
            )

            if run.returncode == 0:
                click.echo()
                click.secho(
                    "Meltano and your Meltano project have been upgraded!",
                    fg="green",
                )
        else:
            click.echo(
                "Then, run `meltano upgrade --skip-package` to upgrade your project based on the latest version."
            )


@upgrade.command(cls=InstrumentedCmd, short_help="Upgrade the Meltano package only.")
@click.option(
    "--pip_url", type=str, envvar="MELTANO_UPGRADE_PIP_URL", help="Meltano pip URL."
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    envvar="MELTANO_UPGRADE_FORCE",
    help="Force upgrade.",
)
@click.pass_context
def package(ctx, **kwargs):
    """Upgrade the Meltano package only."""
    ctx.obj["upgrade_service"].upgrade_package(**kwargs)


@upgrade.command(
    cls=InstrumentedCmd, short_help="Update files managed by file bundles only."
)
@click.pass_context
def files(ctx):
    """Update files managed by file bundles only."""
    ctx.obj["upgrade_service"].update_files()


@upgrade.command(
    cls=InstrumentedCmd, short_help="Apply migrations to system database only."
)
@click.pass_context
def database(ctx):
    """Apply migrations to system database only."""
    ctx.obj["upgrade_service"].migrate_database()



"""Module for managing settings."""
from __future__ import annotations

import logging
import os
import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager
from enum import Enum
from typing import Generator, Iterable

from meltano.core.project import Project
from meltano.core.utils import expand_env_vars as do_expand_env_vars
from meltano.core.utils import flatten

from .setting_definition import SettingDefinition, SettingKind, SettingMissingError
from .settings_store import SettingValueStore

logger = logging.getLogger(__name__)


# sentinel value to use to prevent leaking sensitive data
REDACTED_VALUE = "(redacted)"

# magic string used as feature flag setting for experimental features
EXPERIMENTAL = "experimental"
FEATURE_FLAG_PREFIX = "ff"


class FeatureFlags(Enum):
    """Available Meltano Feature Flags."""

    ENABLE_UVICORN = "enable_uvicorn"
    ENABLE_API_SCHEDULED_JOB_LIST = "enable_api_scheduled_job_list"
    STRICT_ENV_VAR_MODE = "strict_env_var_mode"

    def __str__(self):
        """Return feature name.

        Returns:
            str: Feature name.
        """
        return self.value

    @property
    def setting_name(self) -> str:
        """Return the setting name for this feature flag.

        Returns:
            The setting name for this feature flag.
        """
        return f"{FEATURE_FLAG_PREFIX}.{self.value}"


class FeatureNotAllowedException(Exception):
    """Occurs when a disallowed code path is run."""

    def __init__(self, feature):
        """Instantiate the error.

        Args:
            feature: the feature flag to check
        """
        super().__init__(feature)
        self.feature = feature

    def __str__(self) -> str:
        """Represent the error as a string.

        Returns:
            string representation of the error
        """
        return f"{self.feature} not enabled."


class SettingsService(ABC):  # noqa: WPS214
    """Abstract base class for managing settings."""

    LOGGING = False
    supports_environments = True

    def __init__(
        self,
        project: Project,
        show_hidden: bool = True,
        env_override: dict = None,
        config_override: dict = None,
    ):
        """Create a new settings service object.

        Args:
            project: Meltano project object.
            show_hidden: Whether to display secret setting values.
            env_override: Optional override environment values.
            config_override:  Optional override configuration values.
        """
        self.project = project

        self.show_hidden = show_hidden

        self.env_override = env_override or {}

        self.config_override = config_override or {}

        self._setting_defs = None

    @property
    @abstractmethod
    def label(self):
        """Return label.

        Returns:
            Label for the settings service.
        """

    @property
    @abstractmethod
    def docs_url(self):
        """Return docs URL.

        Returns:
            URL for Meltano doc site.
        """

    @property
    def env_prefixes(self) -> list[str]:
        """Return prefixes for setting environment variables.

        Returns:
            prefixes for settings environment variables
        """
        return ["meltano"]

    @property
    @abstractmethod
    def db_namespace(self) -> str:
        """Return namespace for setting value records in system database."""

    @property
    @abstractmethod
    def setting_definitions(self) -> list[SettingDefinition]:
        """Return definitions of supported settings."""

    @property
    def inherited_settings_service(self):
        """Return settings service to inherit configuration from."""

    @property
    @abstractmethod
    def meltano_yml_config(self) -> dict:
        """Return current configuration in `meltano.yml`."""

    @abstractmethod
    def update_meltano_yml_config(self, config):
        """Update configuration in `meltano.yml`.

        Args:
            config: updated config
        """

    @abstractmethod
    def process_config(self):
        """Process configuration dictionary to be used downstream."""

    @property
    def flat_meltano_yml_config(self):
        """Flatten meltano config.

        Returns:
            the flattened config

        """
        return flatten(self.meltano_yml_config, "dot")

    @property
    def env(self):
        """Return the environment as a dict.

        Returns:
            the environment as a dict.
        """
        return {**os.environ, **self.env_override}

    @classmethod
    def unredact(cls, values: dict) -> dict:
        """Remove any redacted values in a dictionary.

        Args:
            values: the dictionary to remove redacted values from

        Returns:
            the unredacted dictionary
        """
        return {key: val for key, val in values.items() if val != REDACTED_VALUE}

    def config_with_metadata(
        self,
        prefix=None,
        extras=None,
        source=SettingValueStore.AUTO,
        source_manager=None,
        **kwargs,
    ):
        """Return all config values with associated metadata.

        Args:
            prefix: the prefix for setting names
            extras: extra setting definitions to include
            source: the SettingsStore to use
            source_manager: the SettingsStoreManager to use
            kwargs: additional keyword args to pass during SettingsStoreManager instantiation

        Returns:
            dict of config with metadata
        """
        if source_manager:
            source_manager.bulk = True
        else:
            source_manager = source.manager(self, bulk=True, **kwargs)

        config = {}
        for setting_def in self.definitions(extras=extras):
            if prefix and not setting_def.name.startswith(prefix):
                continue

            value, metadata = self.get_with_metadata(
                setting_def.name,
                setting_def=setting_def,
                source=source,
                source_manager=source_manager,
                **kwargs,
            )

            name = setting_def.name
            if prefix:
                name = name[len(prefix) :]  # noqa: E203

            config[name] = {**metadata, "value": value}

        return config

    def as_dict(self, *args, process=False, **kwargs) -> dict:
        """Return settings without associated metadata.

        Args:
            *args: args to pass to config_with_metadata
            process: whether or not to process the config
            **kwargs: additional kwargs to pass to config_with_metadata

        Returns:
            dict of namew-value settings pairs
        """
        config_metadata = self.config_with_metadata(*args, **kwargs)

        if process:
            config = {
                key: metadata["setting"].post_process_value(metadata["value"])
                for key, metadata in config_metadata.items()
            }
            config = self.process_config(config)
        else:
            config = {
                key: metadata["value"] for key, metadata in config_metadata.items()
            }

        return config

    def as_env(self, *args, **kwargs) -> dict[str, str]:
        """Return settings as an dictionary of environment variables.

        Args:
            *args: args to pass to config_with_metadata
            **kwargs: additional kwargs to pass to config_with_metadata

        Returns:
            settings as environment variables
        """
        full_config = self.config_with_metadata(*args, **kwargs)

        env = {}
        for _, config in full_config.items():
            value = config["value"]
            if value is None:
                continue

            setting_def = config["setting"]
            value = setting_def.stringify_value(value)

            for env_var in self.setting_env_vars(setting_def, for_writing=True):
                if env_var.negated:
                    continue

                env[env_var.key] = value

        return env

    def get_with_metadata(  # noqa: WPS210, WPS615
        self,
        name: str,
        redacted=False,
        source=SettingValueStore.AUTO,
        source_manager=None,
        setting_def=None,
        expand_env_vars=True,
        **kwargs,
    ):
        """Get a setting with associated metadata.

        Args:
            name: the name of the setting to get
            redacted: whether or not the setting is redacted
            source: the SettingsStore to use
            source_manager: the SettingsStoreManager to use
            setting_def: get this SettingDefinition instead of name
            expand_env_vars: whether or not to expand nested environment variables
            **kwargs: additional keyword args to pass during SettingsStoreManager instantiation

        Returns:
            a tuple of the setting value and metadata
        """
        try:
            setting_def = setting_def or self.find_setting(name)
        except SettingMissingError:
            pass

        if setting_def:
            name = setting_def.name

        self.log(f"Getting setting '{name}'")

        metadata = {"name": name, "source": source, "setting": setting_def}

        expandable_env = {**self.project.dotenv_env, **self.env}
        if setting_def and setting_def.is_extra:
            expandable_env.update(
                self.as_env(
                    extras=False,
                    redacted=redacted,
                    source=source,
                    source_manager=source_manager,
                )
            )

        manager = source_manager or source.manager(self, **kwargs)
        value, get_metadata = manager.get(name, setting_def=setting_def)
        metadata.update(get_metadata)

        # Can't do conventional SettingsService.feature_flag call to check;
        # it would result in circular dependency
        env_var_strict_mode, _ = manager.get(
            f"{FEATURE_FLAG_PREFIX}.{FeatureFlags.STRICT_ENV_VAR_MODE}"
        )
        if expand_env_vars and metadata.get("expandable", False):
            metadata["expandable"] = False
            expanded_value = do_expand_env_vars(
                value, env=expandable_env, raise_if_missing=env_var_strict_mode
            )

            if expanded_value != value:
                metadata["expanded"] = True
                metadata["unexpanded_value"] = value
                value = expanded_value

        if setting_def:
            # Expand flattened config values if the root value is the default
            # or inherited empty object.
            if setting_def.kind == SettingKind.OBJECT and (
                metadata["source"]
                in {SettingValueStore.DEFAULT, SettingValueStore.INHERITED}
            ):
                object_value = {}
                object_source = metadata["source"]
                for setting_key in [  # noqa: WPS335
                    setting_def.name,
                    *setting_def.aliases,
                ]:
                    flat_config_metadata = self.config_with_metadata(
                        prefix=f"{setting_key}.",
                        redacted=redacted,
                        source=source,
                        source_manager=source_manager,
                        expand_env_vars=expand_env_vars,
                    )
                    for nested_key, config_metadata in flat_config_metadata.items():
                        if nested_key in object_value:
                            continue

                        object_value[nested_key] = config_metadata["value"]

                        nested_source = config_metadata["source"]
                        if nested_source.overrides(object_source):
                            object_source = nested_source

                if object_value:
                    value = object_value
                    metadata["source"] = object_source

            cast_value = setting_def.cast_value(value)
            if cast_value != value:
                metadata["uncast_value"] = value
                value = cast_value

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and setting_def.is_redacted:
                metadata["redacted"] = True
                value = REDACTED_VALUE

        self.log(f"Got setting {name!r} with metadata: {metadata}")

        if setting_def is None and metadata["source"] is SettingValueStore.DEFAULT:
            warnings.warn(
                f"Unknown setting {name!r} - the default value `{value!r}` will be used",
                RuntimeWarning,
            )

        return value, metadata

    def get_with_source(self, *args, **kwargs):
        """Get a setting value along with its source.

        Args:
            *args: args to pass to get_with_metadata
            **kwargs: kwargs to pass to get_with_metadata

        Returns:
            tuple of setting value and its source
        """
        value, metadata = self.get_with_metadata(*args, **kwargs)
        return value, metadata["source"]

    def get(self, *args, **kwargs):
        """Get a setting value.

        Args:
            *args: args to pass to get_with_metadata
            **kwargs: kwargs to pass to get_with_metadata

        Returns:
            the setting value
        """
        value, _ = self.get_with_source(*args, **kwargs)
        return value

    def set_with_metadata(  # noqa: WPS615
        self, path: list[str], value, store=SettingValueStore.AUTO, **kwargs
    ):
        """Set the value and metadata for a setting.

        Args:
            path: the key for the setting
            value: the value to set the setting to
            store: the store to set the value in
            **kwargs: additional keyword args to pass during SettingsStoreManager instantiation

        Returns:
            the new value and metadata for the setting
        """
        self.log(f"Setting setting '{path}'")

        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        metadata = {"name": name, "path": path, "store": store, "setting": setting_def}

        if value == REDACTED_VALUE:
            metadata["redacted"] = True
            return None, metadata

        if setting_def:
            cast_value = setting_def.cast_value(value)
            if cast_value != value:
                metadata["uncast_value"] = value
                value = cast_value

        manager = store.manager(self, **kwargs)
        set_metadata = manager.set(name, path, value, setting_def=setting_def)
        metadata.update(set_metadata)

        self.log(f"Set setting '{name}' with metadata: {metadata}")
        return value, metadata

    def set(self, *args, **kwargs):
        """Set the value for a setting.

        Args:
            *args: args to pass to set_with_metadata
            **kwargs: kwargs to pass to set_with_metadata

        Returns:
            the new value for the setting
        """
        value, _ = self.set_with_metadata(*args, **kwargs)
        return value

    def unset(self, path: list[str], store=SettingValueStore.AUTO, **kwargs):
        """Unset a setting.

        Args:
            path: the key for the setting
            store: the store to set the value in
            **kwargs: additional keyword args to pass during SettingsStoreManager instantiation

        Returns:
            the metadata for the setting
        """
        self.log(f"Unsetting setting '{path}'")

        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        metadata = {"name": name, "path": path, "store": store, "setting": setting_def}

        manager = store.manager(self, **kwargs)
        unset_metadata = manager.unset(name, path, setting_def=setting_def)
        metadata.update(unset_metadata)

        self.log(f"Unset setting '{name}' with metadata: {metadata}")
        return metadata

    def reset(self, store=SettingValueStore.AUTO, **kwargs):
        """Reset a setting.

        Args:
            store: the store to set the value in
            **kwargs: additional keyword args to pass during SettingsStoreManager instantiation

        Returns:
            the metadata for the setting
        """
        metadata = {"store": store}

        manager = store.manager(self, **kwargs)
        reset_metadata = manager.reset()
        metadata.update(reset_metadata)

        self.log(f"Reset settings with metadata: {metadata}")
        return metadata

    def definitions(self, extras=None) -> Iterable[dict]:
        """Return setting definitions along with extras.

        Args:
            extras: additional settings to return

        Returns:
            list of setting definitions
        """
        if self._setting_defs is None:
            self._setting_defs = [
                setting
                for setting in self.setting_definitions
                if setting.kind != SettingKind.HIDDEN or self.show_hidden
            ]

        if extras is not None:
            return [
                setting
                for setting in self._setting_defs
                if (extras is True and setting.is_extra)  # noqa: WPS408
                or (extras is False and not setting.is_extra)
            ]

        return self._setting_defs

    def find_setting(self, name: str) -> SettingDefinition:
        """Find a setting by name.

        Args:
            name:the name or alias of the setting to return

        Returns:
            the setting definition matching the given name

        Raises:
            SettingMissingError: if the setting is not found

        """
        try:
            return next(
                setting
                for setting in self.definitions()
                if setting.name == name or name in setting.aliases
            )
        except StopIteration as err:
            raise SettingMissingError(name) from err

    def setting_env_vars(self, setting_def, for_writing=False):
        """Get environment variables for the given setting definition.

        Args:
            setting_def: the setting definition to get env vars for
            for_writing: unused but referenced elsewhere # TODO: clean up refs at some point

        Returns:
            environment variables for given setting
        """
        return setting_def.env_vars(self.env_prefixes)

    def setting_env(self, setting_def):
        """Get a single environment variable for the given setting definition.

        Args:
            setting_def: the setting definition to get env vars for

        Returns:
            environment variable for given setting
        """
        return self.setting_env_vars(setting_def)[0].key

    def log(self, message):
        """Log the given message.

        Args:
            message: the message to log
        """
        if self.LOGGING:
            logger.debug(message)

    @contextmanager
    def feature_flag(
        self, feature: str, raise_error: bool = True
    ) -> Generator[bool, None, None]:
        """Gate code paths based on feature flags.

        Args:
            feature: the feature flag to check
            raise_error: indicates whether error should be raised

        Yields:
            true if the feature flag is enabled, else false

        Raises:
            FeatureNotAllowedException: if raise_error is True and feature flag is disallowed
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Unknown setting", RuntimeWarning)

            # experimental is a top-level setting
            if feature == EXPERIMENTAL:
                allowed = self.get(EXPERIMENTAL) or False
            # other feature flags are nested under feature flag
            else:
                allowed = self.get(f"{FEATURE_FLAG_PREFIX}.{feature}") or False

        try:
            yield allowed
        finally:
            if raise_error and not allowed:
                raise FeatureNotAllowedException(feature)



"""Install plugins into the project, using pip in separate virtual environments by default."""

from __future__ import annotations

import asyncio
import functools
import logging
import sys
from enum import Enum
from multiprocessing import cpu_count
from typing import Any, Callable, Iterable

from cached_property import cached_property

from meltano.core.plugin.project_plugin import ProjectPlugin

from .error import AsyncSubprocessError, PluginInstallError, PluginInstallWarning
from .plugin import PluginType
from .project import Project
from .project_plugins_service import ProjectPluginsService
from .utils import noop
from .venv_service import VenvService

logger = logging.getLogger(__name__)


class PluginInstallReason(str, Enum):
    """Plugin install reason enum."""

    ADD = "add"
    INSTALL = "install"
    UPGRADE = "upgrade"


class PluginInstallStatus(Enum):
    """The status of the process of installing a plugin."""

    RUNNING = "running"
    SUCCESS = "success"
    SKIPPED = "skipped"
    ERROR = "error"
    WARNING = "warning"


class PluginInstallState:
    """A message reporting the progress of installing a plugin."""

    def __init__(
        self,
        plugin: ProjectPlugin,
        reason: PluginInstallReason,
        status: PluginInstallStatus,
        message: str = None,
        details: str = None,
    ):
        """Initialize PluginInstallState instance.

        Args:
            plugin: Plugin related to this install state.
            reason: Reason for plugin install.
            status: Status of plugin install.
            message: Formatted install state message.
            details: Extra details relating to install (including error details if failed).
        """
        # TODO: use dataclasses.dataclass for this when 3.6 support is dropped
        self.plugin = plugin
        self.reason = reason
        self.status = status
        self.message = message
        self.details = details

    @property
    def successful(self):
        """Plugin install success status.

        Returns:
            'True' if plugin install successful.
        """
        return self.status in {PluginInstallStatus.SUCCESS, PluginInstallStatus.SKIPPED}

    @property
    def skipped(self):
        """Plugin install skipped status.

        Returns:
            'True' if the installation was skipped / not needed.
        """
        return self.status == PluginInstallStatus.SKIPPED

    @property  # noqa: WPS212  # Too many return statements
    def verb(self):
        """Verb form of status.

        Returns:
            Verb form of status.
        """
        if self.status is PluginInstallStatus.RUNNING:
            if self.reason is PluginInstallReason.UPGRADE:
                return "Updating"
            return "Installing"

        if self.status is PluginInstallStatus.SUCCESS:
            if self.reason is PluginInstallReason.UPGRADE:
                return "Updated"

            return "Installed"

        if self.status is PluginInstallStatus.SKIPPED:
            return "Skipped installing"

        return "Errored"


def installer_factory(project, plugin: ProjectPlugin, *args, **kwargs):
    """Installer Factory.

    Args:
        project: Meltano Project.
        plugin: Plugin to be installed.
        args: Positional arguments to instantiate installer with.
        kwargs: Keyword arguments to instantiate installer with.

    Returns:
        An instantiated plugin installer for the given plugin.
    """
    installer_class = PipPluginInstaller
    try:
        installer_class = plugin.installer_class
    except AttributeError:
        pass

    return installer_class(project, plugin, *args, **kwargs)


def with_semaphore(func):
    """Gate access to the method using its class's semaphore.

    Args:
        func: Function to wrap.

    Returns:
        Wrapped function.
    """

    @functools.wraps(func)  # noqa: WPS430
    async def wrapper(self, *args, **kwargs):  # noqa: WPS430
        async with self.semaphore:
            return await func(self, *args, **kwargs)

    return wrapper


class PluginInstallService:
    """Plugin install service."""

    def __init__(
        self,
        project: Project,
        plugins_service: ProjectPluginsService = None,
        status_cb: Callable[[PluginInstallState], Any] = noop,
        parallelism: int | None = None,
        clean: bool = False,
    ):
        """Initialize new PluginInstallService instance.

        Args:
            project: Meltano Project.
            plugins_service: Project plugins service to use.
            status_cb: Status call-back function.
            parallelism: Number of parallel installation processes to use.
            clean: Clean install flag.
        """
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self.status_cb = status_cb
        if parallelism is None:
            self.parallelism = cpu_count()
        elif parallelism < 1:
            self.parallelism = sys.maxsize
        self.clean = clean

    @cached_property
    def semaphore(self):
        return asyncio.Semaphore(self.parallelism)

    @staticmethod
    def remove_duplicates(
        plugins: Iterable[ProjectPlugin], reason: PluginInstallReason
    ):
        """Deduplicate list of plugins, keeping the last occurrences.

        Trying to install multiple plugins into the same venv via `asyncio.run` will fail
        due to a race condition between the duplicate installs. This is particularly
        problematic if `clean` is set as one async `clean` operation causes the other
        install to fail.

        Args:
            plugins: An iterable containing plugins to dedupe.
            reason: Plugins install reason.

        Returns:
            A tuple containing a list of PluginInstallState instance (for skipped plugins)
            and a deduplicated list of plugins to install.
        """
        states = []
        seen_venvs = set()
        deduped_plugins = []
        for plugin in list(plugins):
            if (plugin.type, plugin.venv_name) not in seen_venvs:
                deduped_plugins.append(plugin)
                seen_venvs.add((plugin.type, plugin.venv_name))
            else:
                state = PluginInstallState(
                    plugin=plugin,
                    reason=reason,
                    status=PluginInstallStatus.SKIPPED,
                    message=(
                        f"Plugin '{plugin.name}' does not require installation: "
                        + "reusing parent virtualenv"
                    ),
                )
                states.append(state)
        return states, deduped_plugins

    def install_all_plugins(
        self, reason=PluginInstallReason.INSTALL
    ) -> tuple[PluginInstallState]:
        """
        Install all the plugins for the project.

        Blocks until all plugins are installed.

        Args:
            reason: Plugin install reason.

        Returns:
            Install state of installed plugins.
        """
        return self.install_plugins(self.plugins_service.plugins(), reason=reason)

    def install_plugins(
        self,
        plugins: Iterable[ProjectPlugin],
        reason=PluginInstallReason.INSTALL,
    ) -> tuple[PluginInstallState]:
        """
        Install all the provided plugins.

        Blocks until all plugins are installed.

        Args:
            plugins: ProjectPlugin instances to install.
            reason: Plugin install reason.

        Returns:
            Install state of installed plugins.
        """
        states, new_plugins = self.remove_duplicates(plugins=plugins, reason=reason)
        for state in states:
            self.status_cb(state)
        states.extend(
            asyncio.run(self.install_plugins_async(new_plugins, reason=reason))
        )
        return states

    async def install_plugins_async(
        self,
        plugins: Iterable[ProjectPlugin],
        reason=PluginInstallReason.INSTALL,
    ) -> tuple[PluginInstallState]:
        """Install all the provided plugins.

        Args:
            plugins: ProjectPlugin instances to install.
            reason: Plugin install reason.

        Returns:
            Install state of installed plugins.
        """
        return await asyncio.gather(
            *[self.install_plugin_async(plugin, reason) for plugin in plugins]
        )

    def install_plugin(
        self,
        plugin: ProjectPlugin,
        reason=PluginInstallReason.INSTALL,
    ) -> PluginInstallState:
        """
        Install a plugin.

        Blocks until the plugin is installed.

        Args:
            plugin: ProjectPlugin to install.
            reason: Install reason.

        Returns:
            PluginInstallState state instance.
        """
        return asyncio.run(
            self.install_plugin_async(
                plugin,
                reason=reason,
            )
        )

    @with_semaphore
    async def install_plugin_async(
        self,
        plugin: ProjectPlugin,
        reason=PluginInstallReason.INSTALL,
    ) -> PluginInstallState:
        """Install a plugin asynchronously.

        Args:
            plugin: ProjectPlugin to install.
            reason: Install reason.

        Returns:
            PluginInstallState state instance.
        """
        self.status_cb(
            PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.RUNNING,
            )
        )

        if not plugin.is_installable() or self._is_mapping(plugin):
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.SKIPPED,
                message=f"Plugin '{plugin.name}' does not require installation",
            )
            self.status_cb(state)
            return state

        try:
            async with plugin.trigger_hooks("install", self, plugin, reason):
                await installer_factory(self.project, plugin).install(
                    reason, self.clean
                )
                state = PluginInstallState(
                    plugin=plugin, reason=reason, status=PluginInstallStatus.SUCCESS
                )
                self.status_cb(state)
                return state

        except PluginInstallError as err:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.ERROR,
                message=str(err),
            )
            self.status_cb(state)
            return state

        except PluginInstallWarning as warn:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.WARNING,
                message=str(warn),
            )
            self.status_cb(state)
            return state

        except AsyncSubprocessError as err:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.ERROR,
                message=(
                    f"{plugin.type.descriptor} '{plugin.name}' "
                    + f"could not be installed: {err}"
                ).capitalize(),
                details=await err.stderr,
            )
            self.status_cb(state)
            return state

    @staticmethod
    def _is_mapping(plugin: ProjectPlugin) -> bool:
        """Check if a plugin is a mapping, as mappings are not installed.

        Mappings are PluginType.MAPPERS with extra attribute of `_mapping` which will indicate
        that this instance of the plugin is actually a mapping - and should not be installed.

        Args:
            plugin: ProjectPlugin to evaluate.

        Returns:
            A boolean determining if the given plugin is a mapping (of type PluginType.MAPPERS).
        """
        return plugin.type == PluginType.MAPPERS and plugin.extra_config.get("_mapping")


class PipPluginInstaller:
    """Plugin installer for pip-based plugins."""

    def __init__(
        self,
        project,
        plugin: ProjectPlugin,
        venv_service: VenvService = None,
    ):
        """Initialize PipPluginInstaller instance.

        Args:
            project: Meltano Project.
            plugin: ProjectPlugin to install.
            venv_service: VenvService instance to use when installing.
        """
        self.plugin = plugin
        self.venv_service = venv_service or VenvService(
            project,
            namespace=self.plugin.type,
            name=self.plugin.venv_name,
        )

    async def install(self, reason, clean):
        """Install the plugin into the virtual environment using pip.

        Args:
            reason: Install reason.
            clean: Flag to clean install.

        Returns:
            None.
        """
        return await self.venv_service.install(
            self.plugin.formatted_pip_url, clean=clean
        )



"""A `CommandBlock` pattern supporting Meltano plugin's command like `dbt:run`, `dbt:docs` or `dbt:test`."""

from __future__ import annotations

import asyncio
from abc import ABCMeta, abstractmethod

import structlog

from meltano.core.db import project_engine
from meltano.core.elt_context import PluginContext
from meltano.core.logging import OutputLogger
from meltano.core.logging.utils import SubprocessOutputWriter
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import PluginInvoker, invoker_factory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.runner import RunnerError

from .singer import InvokerBase

logger = structlog.getLogger(__name__)


class PluginCommandBlock(metaclass=ABCMeta):
    """Basic PluginCommand interface specification."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the plugin command block.

        In the case of a singer plugin this would most likely be the name of the plugin.
        ex. `dbt:run` name = dbt , command = run
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def command(self) -> str | None:
        """Command is the specific plugin command to use when invoking the plugin (if any)."""
        raise NotImplementedError

    @abstractmethod
    async def run(self) -> None:
        """Run the command."""
        raise NotImplementedError


class InvokerCommand(InvokerBase, PluginCommandBlock):
    """A basic PluginCommandBlock interface implementation that supports running plugin commands."""

    def __init__(
        self,
        name: str,
        log: SubprocessOutputWriter,
        block_ctx: dict,
        project: Project,
        plugins_service: ProjectPluginsService,
        plugin_invoker: PluginInvoker,
        command: str | None,
        command_args: tuple[str],
    ):
        """Configure and return a wrapped plugin invoker.

        Args:
            name: the name of the plugin/command.
            log: the OutputLogger instance to proxy output too.
            block_ctx: the block context.
            project: the project instance.
            plugins_service: the project plugins service.
            plugin_invoker: the plugin invoker.
            command: the command to invoke.
            command_args: any additional plugin args that should be used.
        """
        super().__init__(
            block_ctx=block_ctx,
            project=project,
            plugins_service=plugins_service,
            plugin_invoker=plugin_invoker,
            command=command,
        )
        self._name = name
        self._command = command
        self._command_args = command_args
        self._log = log

    @property
    def name(self) -> str:
        """Name is the underlying name of the plugin/command.

        Returns:
            The name str.
        """
        return self._name

    @property
    def command(self) -> str | None:
        """Command is the specific plugin command to use when invoking the plugin.

        Returns:
            The command str if any.
        """
        return self._command

    @property
    def command_args(self) -> str | None:
        """Command args are the specific plugin command args to use when invoking the plugin (if any).

        Returns:
            The command args if any.
        """
        return self._command_args

    async def _start(self):
        invoke_args = (self.command_args,) if self.command_args else ()
        await self.start(*invoke_args)

    async def run(self) -> None:
        """Invoke a command capturing and logging produced output.

        Raises:
            RunnerError: if the command fails.
        """
        try:  # noqa: WPS501
            async with self.invoker.prepared(self.context.session):
                await self._start()

                self.stdout_link(self._log)
                self.stderr_link(self._log)

                await asyncio.wait(
                    [*self.proxy_io(), self.process_future],
                    return_when=asyncio.ALL_COMPLETED,
                )
        finally:
            self.context.session.close()
        exitcode = self.process_future.result()
        if exitcode:
            command = self.command or self.command_args[0]
            raise RunnerError(
                f"`{self.name} {command}` failed with exit code: {exitcode}"
            )


def plugin_command_invoker(
    plugin: ProjectPlugin,
    project: Project,
    command: str | None,
    command_args: list[str] | None = None,
    run_dir: str = None,
) -> InvokerCommand:
    """
    Make an InvokerCommand from a plugin.

    Args:
        plugin: Plugin to make command from.
        project: Project to use.
        command: the command to invoke on the plugin i.e. `run` in dbt run.
        command_args: any additional command args that should be passed in during invocation.
        run_dir: Optional directory to run commands in.

    Returns:
        InvokerCommand
    """
    stderr_log = logger.bind(
        stdio="stderr",
        cmd_type="command",
    )

    _, session_maker = project_engine(project)
    session = session_maker()

    output_logger = OutputLogger("run.log")
    invoker_log = output_logger.out(plugin.name, stderr_log)

    plugins_service = ProjectPluginsService(project)

    ctx = PluginContext(
        plugin=plugin,
        settings_service=PluginSettingsService(
            project,
            plugin,
            plugins_service=plugins_service,
        ),
        session=session,
    )

    invoker = invoker_factory(
        project,
        ctx.plugin,
        context=ctx,
        run_dir=run_dir,
        plugins_service=plugins_service,
        plugin_settings_service=ctx.settings_service,
    )
    return InvokerCommand(
        name=plugin.name,
        log=invoker_log,
        block_ctx=ctx,
        project=project,
        plugins_service=plugins_service,
        plugin_invoker=invoker,
        command=command,
        command_args=command_args,
    )



"""This holds the actual BlockSet meta class as well as related components such as exceptions."""
from __future__ import annotations

from abc import ABCMeta, abstractmethod


class BlockSetValidationError(Exception):
    """Base exception when a block in a BlockSet violates the sets requirements."""

    def __init__(self, error: str, message: str = "block violates set requirements"):
        """Initialize exception for when plugin violates the BlockSet requirements.

        Args:
            error: The error.
            message: The message.
        """
        super().__init__(f"{message}: {error}")


class BlockSet(metaclass=ABCMeta):
    """Currently the only complex block set is our ExtractLoadBlocks type.

    Theoretically, this is the bare minimum that we need to run and terminate (i.e. early abort) a block set. So anything
    implementing a run(), terminate(), and validate_set() method currently satisfies the BlockSet interface.
    """

    @abstractmethod
    async def run(self) -> None:
        """Do whatever a BlockSet is designed to do."""
        raise NotImplementedError

    @abstractmethod
    async def terminate(self, graceful: bool = True):
        """Terminate a currently executing BlockSet.

        Args:
            graceful: Whether the BlockSet should try to gracefully quit.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_set(self) -> None:
        """Validate a block set to ensure its valid and runnable.

        Raises:
            BlockSetValidationError: on validation failure
        """
        raise NotImplementedError



"""Service to manage meltano.yml."""
from __future__ import annotations

import logging
import os
from contextlib import contextmanager

import yaml

from meltano.core import bundle
from meltano.core.project import Project
from meltano.core.setting_definition import SettingDefinition

logger = logging.getLogger(__name__)


class ConfigService:
    """Service to manage meltano.yml."""

    def __init__(self, project: Project, use_cache=True):
        """Create a new project configuration service.

        Args:
            project: the project to configure.
            use_cache: whether to use the cache or not.
        """
        self.project = project

        self._settings = None
        self._current_meltano_yml = None
        self._use_cache = use_cache

    @property
    def settings(self):
        """Return the project settings.

        Returns:
            The project settings.
        """
        if self._settings is None:
            with open(bundle.root / "settings.yml") as settings_yaml:
                settings = yaml.safe_load(settings_yaml)
            self._settings = list(map(SettingDefinition.parse, settings["settings"]))

        return self._settings

    @property
    def current_meltano_yml(self):
        """Return the current meltano.yml contents.

        Returns:
            The contents of meltano.yml.
        """
        if self._current_meltano_yml is None or not self._use_cache:
            self.project.clear_cache()
            self._current_meltano_yml = self.project.meltano
        return self._current_meltano_yml

    @contextmanager
    def update_meltano_yml(self):
        """Update meltano.yml.

        This method is a context manager that will update meltano.yml with the
        contents of the context.

        Yields:
            The contents of meltano.yml.
        """
        with self.project.meltano_update() as meltano_yml:
            yield meltano_yml

        self._current_meltano_yml = None

    @contextmanager
    def update_active_environment(self):
        """Update active environment.

        Yields:
            active environment
        """
        environment = self.project.active_environment

        with self.update_meltano_yml() as meltano_yml:
            environments = meltano_yml.environments

            # find the proper environment to update
            env_idx, _ = next(
                (idx, env) for idx, env in enumerate(environments) if env == environment
            )

            active_environment = environments[env_idx]
            yield active_environment

        self.project.active_environment = active_environment

    @property
    def current_config(self):
        """Return the current configuration.

        Returns:
            The current configuration.
        """
        return self.current_meltano_yml.extras

    def update_config(self, config):
        """Update top-level Meltano configuration.

        Args:
            config: configuration dict
        """
        with self.update_meltano_yml() as meltano_yml:
            meltano_yml.extras = config

    def make_meltano_secret_dir(self):
        """Create the secret directory."""
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    @property
    def env(self):
        """Return the top-level env vars from meltano.yml.

        Returns:
            A dictionary of (unexpanded) environment variables.
        """
        return self.current_meltano_yml.env



"""Settings manager for Meltano plugins."""

from __future__ import annotations

from typing import Any

from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.setting_definition import SettingDefinition
from meltano.core.settings_service import FeatureFlags, SettingsService
from meltano.core.utils import expand_env_vars


class PluginSettingsService(SettingsService):
    """Settings manager for Meltano plugins."""

    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        *args,
        plugins_service: ProjectPluginsService = None,
        **kwargs,
    ):
        """Create a new plugin settings manager.

        Args:
            project: The Meltano project.
            plugin: The Meltano plugin.
            args: Positional arguments to pass to the superclass.
            plugins_service: The Meltano plugins service.
            kwargs: Keyword arguments to pass to the superclass.
        """
        super().__init__(project, *args, **kwargs)

        self.plugin = plugin
        self.plugins_service = plugins_service or ProjectPluginsService(self.project)

        self._inherited_settings_service = None
        if self.project.active_environment:
            environment = self.project.active_environment
            self.environment_plugin_config = environment.get_plugin_config(
                self.plugin.type,
                self.plugin.name,
            )
        else:
            self.environment_plugin_config = None

        project_settings_service = ProjectSettingsService(
            self.project, config_service=self.plugins_service.config_service
        )

        self.env_override = {
            **project_settings_service.env,  # project level environment variables
            **project_settings_service.as_env(),  # project level settings as env vars (e.g. MELTANO_PROJECT_ID)
            **self.env_override,  # plugin level overrides, passed in as **kwargs and set to self.env_overrides by super().__init__ above
            **self.plugin.info_env,  # generated generic plugin settings as env vars (e.g. MELTANO_EXTRACT_NAME)
            **self.plugin.env,  # env vars stored under the `env:` key of the plugin definition
        }

        environment_env = {}
        if self.project.active_environment:
            with self.feature_flag(
                FeatureFlags.STRICT_ENV_VAR_MODE, raise_error=False
            ) as strict_env_var_mode:
                environment_env = {
                    var: expand_env_vars(
                        value,
                        self.env_override,
                        raise_if_missing=strict_env_var_mode,
                    )
                    for var, value in self.project.active_environment.env.items()
                }
            self.env_override.update(
                environment_env
            )  # active Meltano Environment top level `env:` key

        environment_plugin_env = (
            self.environment_plugin_config.env if self.environment_plugin_config else {}
        )
        self.env_override.update(
            environment_plugin_env
        )  # env vars stored under the `env:` key of the plugin definition of the active meltano Environment

    @property
    def label(self):
        """Get the label for this plugin.

        Returns:
            The label for this plugin.
        """
        return f"{self.plugin.type.descriptor} '{self.plugin.name}'"  # noqa: WPS237

    @property
    def docs_url(self):
        """Get the documentation URL for this plugin.

        Returns:
            The documentation URL for this plugin.
        """
        return self.plugin.docs

    def setting_env_vars(self, setting_def: SettingDefinition, for_writing=False):
        """Get environment variables for a setting.

        Args:
            setting_def: The setting definition.
            for_writing: Whether to get environment variables for writing.

        Returns:
            Environment variables for a setting.
        """
        return setting_def.env_vars(
            prefixes=self.plugin.env_prefixes(for_writing=for_writing),
            include_custom=self.plugin.is_shadowing or for_writing,
            for_writing=for_writing,
        )

    @property
    def db_namespace(self):
        """Return namespace for setting value records in system database.

        Returns:
            Namespace for setting value records in system database.
        """
        # "default" is included for legacy reasons
        return ".".join((self.plugin.type, self.plugin.name, "default"))

    @property
    def setting_definitions(self) -> list[SettingDefinition]:
        """Return definitions of supported settings.

        Returns:
            A list of setting definitions.
        """
        settings = self.plugin.settings_with_extras

        if self.environment_plugin_config is not None:
            settings.extend(
                self.environment_plugin_config.get_orphan_settings(settings)
            )

        return settings

    @property
    def meltano_yml_config(self):
        """Return current configuration in `meltano.yml`.

        Returns:
            Current configuration in `meltano.yml`.
        """
        return self.plugin.config_with_extras

    @property
    def environment_config(self):
        """Return current environment configuration in `meltano.yml`.

        Returns:
            Current environment configuration in `meltano.yml`.
        """
        if self.environment_plugin_config:
            return self.environment_plugin_config.config_with_extras
        return {}

    def update_meltano_yml_config(self, config_with_extras):
        """Update configuration in `meltano.yml`.

        Args:
            config_with_extras: Configuration to update.
        """
        self.plugin.config_with_extras = config_with_extras
        self.plugins_service.update_plugin(self.plugin)

    def update_meltano_environment_config(self, config_with_extras: dict[str, Any]):
        """Update environment configuration in `meltano.yml`.

        Args:
            config_with_extras: Configuration to update.
        """
        self.environment_plugin_config.config_with_extras = config_with_extras
        self.plugins_service.update_environment_plugin(self.environment_plugin_config)

    @property
    def inherited_settings_service(self):
        """Return settings service to inherit configuration from.

        Returns:
            Settings service to inherit configuration from.
        """
        parent_plugin = self.plugin.parent
        if not isinstance(parent_plugin, ProjectPlugin):
            return None

        if self._inherited_settings_service is None:
            self._inherited_settings_service = self.__class__(
                self.project,
                parent_plugin,
                env_override=self.env_override,
                plugins_service=self.plugins_service,
            )
        return self._inherited_settings_service

    def process_config(self, config):
        """Process configuration dictionary to be passed to plugin.

        Args:
            config: Configuration dictionary to process.

        Returns:
            Processed configuration dictionary.
        """
        return self.plugin.process_config(config)



"""Meltano Canonical object behavior."""

from __future__ import annotations

import copy
import json
from functools import lru_cache
from os import PathLike
from typing import Any, TypeVar

import yaml
from ruamel.yaml import Representer
from ruamel.yaml.comments import CommentedMap, CommentedSeq, CommentedSet

T = TypeVar("T", bound="Canonical")  # noqa: WPS111 (name too short)


class IdHashBox:
    """Wrapper class that makes the hash of an object its Python ID."""

    def __init__(self, content: Any):
        """Initialize the `IdHashBox`.

        Parameters:
            content: The object that will be stored within the `IdHashBox`.
                Its Python ID will be used to hash the `IdHashBox` instance.
        """
        self.content = content

    def __hash__(self) -> int:
        """Hash the `IdHashBox` according to the Python ID of its content.

        Returns:
            The Python ID of the content of the `IdHashBox` instance.
        """
        return id(self.content)

    def __eq__(self, other: Any) -> bool:
        """Check equality of this instance and some other object.

        Parameters:
            other: The object to check equality with.

        Returns:
            Whether this instance and `other` have the same hash.
        """
        return hash(self) == hash(other)


CANONICAL_PARSE_CACHE_SIZE = 4096


class Canonical:  # noqa: WPS214 (too many methods)
    """Defines an object that can be represented as a subset of its attributes.

    Its purpose is to be serializable as the smallest possible form.

    The attribute rules are:
      - All attributes that are Truthy
      - All attributes that are boolean (False is valid)
      - All attributes that are listed in the `_verbatim` set and non-null
      - All attributes that start with "_" are excluded
    """

    def __init__(self, *args: Any, **attrs: Any):
        """Initialize the current instance with the given attributes.

        Args:
            args: Arguments to initialize with.
            attrs: Keyword arguments to initialize with.
        """
        self._dict = CommentedMap()

        super().__init__(*args)

        for attr, value in attrs.items():
            setattr(self, attr, value)

        self._verbatim = set()
        self._flattened = {"extras"}

        self._fallback_to = None
        self._fallbacks = set()
        self._defaults = {}

    @classmethod
    def as_canonical(
        cls: type[T], target: Any
    ) -> dict | list | CommentedMap | CommentedSeq | Any:
        """Return a canonical representation of the given instance.

        Args:
            target: Instance to convert.

        Returns:
            Canonical representation of the given instance.
        """
        if isinstance(target, Canonical):
            result = CommentedMap([(key, cls.as_canonical(val)) for key, val in target])
            target.attrs.copy_attributes(result)
            return result

        if isinstance(target, (CommentedSet, CommentedSeq)):
            result = CommentedSeq(cls.as_canonical(val) for val in target)
            target.copy_attributes(result)
            return result

        if isinstance(target, CommentedMap):
            results = CommentedMap()
            for key, val in target.items():
                if isinstance(val, Canonical):
                    results[key] = val.canonical()
                else:
                    results[key] = cls.as_canonical(val)
            target.copy_attributes(results)
            return results

        if isinstance(target, (list, set)):
            return list(map(cls.as_canonical, target))

        if isinstance(target, dict):
            return {
                key: val.canonical()
                if isinstance(val, Canonical)
                else cls.as_canonical(val)
                for key, val in target.items()
            }

        return copy.deepcopy(target)

    def canonical(self) -> dict | list | CommentedMap | CommentedSeq | Any:
        """Return a canonical representation of the current instance.

        Returns:
            A canonical representation of the current instance.
        """
        return type(self).as_canonical(self)

    def with_attrs(self: T, *args: Any, **kwargs: Any) -> T:
        """Return a new instance with the given attributes set.

        Args:
            args: Attribute names to set.
            kwargs: Keyword arguments to set.

        Returns:
            A new instance with the given attributes set.
        """
        return type(self)(**{**self.canonical(), **kwargs})

    @classmethod
    def parse(cls: type[T], obj: Any) -> T:
        """Parse a 'Canonical' object from a dictionary or return the instance.

        Args:
            obj: Dictionary or instance to parse.

        Returns:
            Parsed instance.
        """
        return cls._parse(IdHashBox(obj))

    @classmethod
    @lru_cache(maxsize=CANONICAL_PARSE_CACHE_SIZE)
    def _parse(cls: type[T], boxed_obj: IdHashBox) -> T:
        """Parse a `Canonical` object from a dictionary or return the instance.

        Args:
            boxed_obj: Dictionary or instance to parse wrapped in an `IdHashBox`.

        Returns:
            Parsed instance.
        """
        obj = boxed_obj.content

        if obj is None:
            return None

        if isinstance(obj, cls):
            return obj

        instance = cls(**obj)

        if isinstance(obj, CommentedMap):
            obj.copy_attributes(instance.attrs)

        return instance

    @property
    def attrs(self) -> CommentedMap:
        """Return the attributes of the current instance.

        Returns:
            Attributes of the current instance.
        """
        return self._dict

    def is_attr_set(self, attr):
        """Return whether specified attribute has a non-default/fallback value set.

        Args:
            attr: Attribute to check.

        Returns:
            True if attribute has a non-default/fallback value set.
        """
        return self._dict.get(attr) is not None

    def __getattr__(self, attr: str) -> Any:
        """Return the value of the given attribute.

        Args:
            attr: Attribute to return.

        Returns:
            Value of the given attribute.

        Raises:
            AttributeError: If the attribute is not set.
        """
        try:
            value = self._dict[attr]
        except KeyError as err:
            if self._fallback_to and not attr.startswith("_"):
                return getattr(self._fallback_to, attr)

            raise AttributeError(attr) from err

        if value is not None:
            return value

        if attr in self._fallbacks and self._fallback_to:
            value = getattr(self._fallback_to, attr)

        if value is not None:
            return value

        if attr in self._defaults:
            value = self._defaults[attr](self)  # noqa: WPS529 (implicit .get())

        return value

    def __setattr__(self, attr: str, value: Any):
        """Set the given attribute to the given value.

        Args:
            attr: Attribute to set.
            value: Value to set.
        """
        if attr.startswith("_") or hasattr(type(self), attr):  # noqa: WPS421
            super().__setattr__(attr, value)
        else:
            self._dict[attr] = value

    def __getitem__(self, attr: str) -> Any:
        """Return the value of the given attribute.

        Args:
            attr: Attribute to return.

        Returns:
            Value of the given attribute.
        """
        return getattr(self, attr)

    def __setitem__(self, attr: str, value: Any) -> None:
        """Set the given attribute to the given value.

        Args:
            attr: Attribute to set.
            value: Value to set.

        Returns:
            None.
        """
        return setattr(self, attr, value)

    def __iter__(self):
        """Return an iterator over the attributes set on the current instance.

        Yields:
            An iterator over the attributes set on the current instance.
        """
        for key, val in self._dict.items():
            if not val:
                if key in self._verbatim:
                    if val is None:
                        continue
                else:
                    # bool values are valid and should be forwarded
                    if val is not False:
                        continue

            # empty canonicals should be skipped
            if isinstance(val, Canonical) and not dict(val):
                continue

            if key in self._flattened:
                if isinstance(val, Canonical):
                    yield from val
                else:
                    yield from val.items()
            else:
                yield (key, val)

    def __len__(self):
        """Return the number of attributes set on the current instance.

        Returns:
            Number of attributes set on the current instance.
        """
        return len(self._dict)

    def __contains__(self, obj: Any):
        """Return whether the current instance contains the given object.

        Args:
            obj: Object to check.

        Returns:
            True if the current instance contains the given object.
        """
        return obj in self._dict

    def update(self, *others: Any, **kwargs: Any) -> None:
        """Update the current instance with the given others.

        Args:
            others: Other instances to update with.
            kwargs: Keyword arguments to update with.
        """
        if kwargs:
            others = [*others, kwargs]

        for other in others:
            other = type(self).as_canonical(other)

            for key, val in other.items():
                setattr(self, key, val)

    @classmethod
    def yaml(cls, dumper: yaml.BaseDumper, obj: Any) -> yaml.MappingNode:
        """YAML serializer for Canonical objects.

        Args:
            dumper: The YAML dumper.
            obj: The Canonical object to serialize.

        Returns:
            The serialized YAML representation of the object.
        """
        return dumper.represent_mapping(
            "tag:yaml.org,2002:map", cls.as_canonical(obj), flow_style=False
        )

    @classmethod
    def to_yaml(cls, representer: Representer, obj: Any):
        """YAML serializer for Canonical objects.

        Args:
            representer: The YAML representer.
            obj: The Canonical object to serialize.

        Returns:
            The serialized YAML representation of the object.
        """
        return representer.represent_mapping(
            "tag:yaml.org,2002:map", cls.as_canonical(obj)
        )

    @classmethod
    def parse_json_file(cls: type[T], path: PathLike) -> T:
        """Parse a plugin definition from a JSON file.

        Args:
            path: The path to the JSON file.

        Returns:
            A standalone plugin definition.
        """
        with open(path) as file:
            return cls.parse(json.load(file))


yaml.add_multi_representer(Canonical, Canonical.yaml)
